#!/usr/local/sbin/charm-env python3
#
# Copyright 2016,2019 Canonical Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Define hw-health charm actions."""

import os
import subprocess
import sys
from datetime import datetime

from charmhelpers.core import unitdata
from charmhelpers.core.hookenv import action_fail, action_get, action_set, config, log

from hwhealth import tools


IPMI_SEL = "/usr/sbin/ipmi-sel"


def clear_sel():
    """Clear the IPMI system event log.

    Implementation uses 'ipmi-sel --post-clear'
    Clears the SEL log and stores the cleared entries in action output.
    """
    command = [IPMI_SEL, "--post-clear"]
    try:
        output = subprocess.check_output(command)
        log("Action clear-sel completed, sel log cleared: {}".format(output))
        action_set({"message": output.decode("UTF-8")})
    except subprocess.CalledProcessError as e:
        action_fail("Action failed with {}".format(e))


def show_sel():
    """Show the entries in the IPMI system event log.

    By default, this will show all non-nominal events.
    If you specify show-all, it will show all events.
    """
    show_all = action_get("show-all")
    command = [IPMI_SEL, "--output-event-state"]

    date_start = unitdata.kv().get(tools.Ipmi.SEL_DATE_FILTER_START_KEY)
    if date_start:
        command.extend(["--date-range", "{}-now".format(date_start)])

    try:
        header, body = None, None
        output = subprocess.check_output(command).decode("UTF-8")
        lines = output.splitlines()
        if lines:
            header, body = lines[0], lines[1:]
            if not show_all:
                # This is fairly naive, but it may be good enough for now
                body = [line for line in body if "Nominal" not in line]
        if body:
            final_output = "\n".join([header] + body)
            if date_start:
                final_output += "Has been date filtered: {}".format(date_start)
        else:
            final_output = "No matching entries found"
        log("Action show-sel completed:\n{}".format(final_output))
        action_set({"message": final_output})
    except subprocess.CalledProcessError as e:
        action_fail("Action failed with {}".format(e))


def ack_sel():
    """Set the acknowledgement date filter."""
    message = ""
    if "--date-range" in config("ipmi_check_options"):
        message = (
            "WARNING: Will override user specified --date-range "
            "option on ipmi_check_options config\n"
        )

    date_param = action_get("date")
    if date_param:
        try:
            date = datetime.strptime(date_param, "%Y-%m-%d")
        except ValueError:
            action_fail(
                "Invalid date format. Expected format is ISO 8601 "
                "(YYYY-MM-DD), got: {}".format(date_param)
            )
            return
    else:
        date = datetime.utcnow()

    try:
        unitdata.kv().set(
            tools.Ipmi.SEL_DATE_FILTER_START_KEY, date.strftime("%m/%d/%Y")
        )
        unitdata.kv().flush()
        message += "Only SEL entries newer than {} will be considered".format(
            date.strftime("%Y-%m-%d")
        )
        action_set({"message": message})
        tools.Ipmi().install_cronjob()
    except Exception as e:
        action_fail("Action failed with {}".format(e))


def unack_sel():
    """Unset the acknowledgement date filter."""
    try:
        unitdata.kv().unset(tools.Ipmi.SEL_DATE_FILTER_START_KEY)
        unitdata.kv().flush()
        message = "SEL entries from all times will be considered"
        action_set({"message": message})
        tools.Ipmi().install_cronjob()
    except Exception as e:
        action_fail("Action failed with {}".format(e))


# A dictionary of all the defined actions to callables (which take
# parsed arguments).
ACTIONS = {
    "clear-sel": clear_sel,
    "show-sel": show_sel,
    "ack-sel": ack_sel,
    "unack-sel": unack_sel,
}


def main(args):
    """Define main subroutine."""
    action_name = os.path.basename(args[0])

    try:
        ACTIONS[action_name]()
    except Exception as e:
        action_fail(str(e))


if __name__ == "__main__":
    sys.exit(main(sys.argv))
