# -----------------------------------------------
# This file is juju managed
# -----------------------------------------------

# Copyright (C) 2008-2013  James Troup <james.troup@canonical.com>
# Copyright (c) 2020 Drew Freiberger <drew.freiberger@canonical.com>
"""Implement a library of functions common to several hw-health scripts."""

import argparse
import os
import re
import subprocess
import tempfile
from binascii import crc32
from datetime import datetime, timedelta
from socket import gethostname


import yaml

ignoring_output = []
_SSACLI_ALT_BIN = [
    "/usr/sbin/hpssacli",
    "/opt/smartstorageadmin/ssacli/bin/ssacli",
]
SSACLI_BIN = next(
    (f for f in _SSACLI_ALT_BIN if os.path.isfile(f) and os.access(f, os.X_OK)),
    "/usr/sbin/ssacli",
)


def read_ignore_file(ignore_file):  # noqa C901
    """Create a list of ignore dicts.

    Reads a file to create the dicts which are used
    with the ignore function.

    The directory /etc/nagios/ignores/ must be owned by nagios user,
    so that it can create and re-write the ignores files.
    Use the following to create it with correct perms:
      D=/etc/nagios/ignores; sudo mkdir $D; sudo chown nagios $D;
      MODULE=power # or array, controller, fans
      sudo -u nagios vim /etc/nagios/ignores/ignores-cat-hp-$MODULE.txt

    The following is useful to paste at the top of an ignores file:
    # This file used by /usr/lib/nagios/plugins/check_hp_hardware.py
    # so that specific hardware errors can be ignored
    # and the nagios alert NOT downtimed or acknowledged,
    # thus allowing other hardware failures to alert.
    # ----------------------------------------------------------------
    # Ignore lines are an expiry date and a substring to match, like:
    # [YYYY-mm-dd HH:MM] nagios error string
    # Lines will be disabled when not matched so they can alert again,
    # but starting a line with '*' will persist a line when not matched
    # and should be used for flapping alerts/hardware only.
    # Comment lines (starting with '#') and blank lines are ignored.
    # Example ignore lines:
    # # do not alert on fan 1 until 2016-11-17 11:00
    # [2016-11-17 11:00] System Board (1): fans are not redundant
    # # ignore power supply 1 until 2016-11-17 0:00, even if unmatched
    # * [2016-11-17] Pwr. Supply Bay (1): power supply is 'Failed'

    # example ignore dict
    # {'matched': False, # set to True if it matches, check before write
    #  'expired': False, # has ignore date passed
    #  'line': "[2016-11-01 23:00] substring to match"
    #  'ignore': "substring to match OR None", # None if expired
    # }
    """
    ignores = []
    if os.path.isfile(ignore_file):
        for line in open(ignore_file).readlines():
            d = {
                "matched": False,
                "expired": False,
                "line": line.rstrip(),
                "ignore": None,
            }
            line = line.strip()
            # special case lines starting with '*', do not disable if unmatched
            persist = False
            if line.startswith("*"):
                persist = True
                line = line.lstrip("*").strip()
            # parse date lines
            if line.startswith("["):
                parts = re.split("\\[|\\]", line, maxsplit=2)
                date = parts[1].strip()
                ignore = parts[2].strip()
                try:
                    date = datetime.strptime(date, "%Y-%m-%d %H:%M")
                except ValueError:
                    try:
                        date = datetime.strptime(date, "%Y-%m-%d")
                    except ValueError:
                        print("Failed to parse ignore date: {}".format(d["line"]))
                        date = None
                    if date:
                        # Do not alert directly at midnight UTC
                        host_crc = crc32(gethostname()) & 0xFFFFFFFF
                        date += timedelta(seconds=(host_crc % 86400))
                # check if date valid and is in the future
                if date:
                    if datetime.now().weekday() in (5, 6):
                        # Ignore Saturday/Sunday to not annoy on-call
                        d["ignore"] = ignore
                    elif date > datetime.now():
                        d["ignore"] = ignore
                    else:
                        d["expired"] = True
                    if persist and not d["expired"]:
                        # set matched True so will not get disabled on non-match
                        d["matched"] = True
            # comment lines and empty lines are just added
            elif line.startswith("#") or not line:
                # add with matched True so does not trigger a file rewrite
                d["matched"] = True
            # unrecognized lines left matched False so rewritten
            ignores.append(d)
    return ignores


def write_ignore_file(ignores, ignore_file):
    """Re-process and save the exclude_file."""
    # if any ignores are not matched then write out file lines again
    if any([not i["matched"] for i in ignores]):
        dirname, basename = os.path.split(ignore_file)
        date = datetime.now()
        try:
            f = tempfile.NamedTemporaryFile(dir=dirname, prefix=basename, delete=False)
            for ignore in ignores:
                if not ignore["matched"] and ignore["ignore"]:
                    ignore["line"] = "# not matched at {} #{}".format(
                        date.strftime("%Y-%m-%dT%H:%M:%S"), ignore["line"]
                    )
                elif ignore["expired"]:
                    # this won't get updated unless the alert has cleared
                    ignore["line"] = "# expired #{}".format(ignore["line"])
                elif not ignore["matched"]:
                    ignore["line"] = "# unknown or bad format #{}".format(
                        ignore["line"]
                    )
                f.write(ignore["line"] + "\n")
            f.flush()
            os.fsync(f.fileno())
            f.close()
            # NamedTemporaryFile has restricted permissions, open them up a bit
            os.chmod(f.name, os.stat(f.name).st_mode | 0o044)
            os.rename(f.name, ignore_file)
        except OSError:
            print("Failed to write {}".format(ignore_file))
            pass


def ignore(line, ignores):
    """Determine if line matches any exclude_file entry."""
    for ignore in ignores:
        if ignore["ignore"] and ignore["ignore"] in line:
            ignoring_output.append("Ignoring: {}".format(line))
            # note: ignores can be updated since it is passed by reference
            # matched=True to keep using this ignore (see write_ignore_file)
            ignore["matched"] = True
            return True
    return False


def get_hp_controller_slots():
    """Determine slot of HP storage controller.

    Use the utility to determine the current controller slot(s) available for probing
    """
    slots = []
    cmd = [SSACLI_BIN, "ctrl", "all", "show"]
    try:
        results = subprocess.check_output(cmd).decode("UTF-8")
    except subprocess.CalledProcessError:
        return slots
    for line in results.splitlines():
        if "in Slot" in line:
            slots.append(line.split()[5])
    return slots


class HWCronArgumentParser(argparse.ArgumentParser):
    """Hardware health Cron job argument parser.

    Provides argparse flags common to most hw-health check cronjobs.
    """

    def __init__(self, def_write_file=None, *args, **kwargs):
        """Initialize argparse defaults for cronjobs based on script name."""
        super().__init__(
            formatter_class=argparse.ArgumentDefaultsHelpFormatter, *args, **kwargs
        )
        # self.prog is populated by ArgumentParser
        self._def_write_file = def_write_file or "/var/lib/nagios/{}.out".format(
            self.prog
        )

    def parse_args(self, *args, **kwargs):
        """Process common CLI args for cron scripts."""
        self.add_argument(
            "-w",
            "--write",
            dest="write",
            type=str,
            default=self._def_write_file,
            help="cache tool output in this file",
        )
        super().parse_args(*args, **kwargs)


class HWCheckArgumentParser(argparse.ArgumentParser):
    """Hardware health Nagios check argument parser.

    Provides argparse flags common to most hw-health nagios check scripts.
    """

    def __init__(self, def_input_file=None, *args, **kwargs):
        """Initialize argparse defaults for nagios checks based on script name."""
        super().__init__(
            formatter_class=argparse.ArgumentDefaultsHelpFormatter, *args, **kwargs
        )
        # self.prog is populated by ArgumentParser
        self._def_input_file = def_input_file or "/var/lib/nagios/{}.out".format(
            self.prog
        )

    def parse_args(self, *args, **kwargs):
        """Process common CLI args for nagios check scripts."""
        self.add_argument(
            "-i",
            "--input",
            dest="input_file",
            type=str,
            default=self._def_input_file,
            help="read cached tool output from this file",
        )
        super().parse_args(*args, **kwargs)


class HPArgumentParser(HWCronArgumentParser):
    """Hardware health Nagios check argument parser.

    Provides argparse flags common to most hp vendor health scripts.
    This class of ArgParse includes exclude_file/ignore managment in order to
    silence known/long-running issues without disabling the entire check.
    """

    def __init__(self, def_exclude_file=None, *args, **kwargs):
        """Initialize argparse exclude_file defaults for HP nagios tools."""
        super().__init__(*args, **kwargs)
        self._def_exclude_file = (
            def_exclude_file or "/etc/nagios/{}.exclude.yaml".format(self.prog)
        )

    def _expired(self, exclusion):
        """Check if we are now past the expiration of an exlude_file entry."""
        return "expires" in exclusion and exclusion["expires"] < datetime.now()

    def parse_args(self, *args, **kwargs):
        """Process common CLI args for HP nagios check scripts."""
        self.add_argument(
            "--debug",
            dest="debug",
            action="store_true",
            help="Extra debugging",
        )

        self.add_argument(
            "--exclude",
            dest="exclude",
            type=str,
            action="append",
            help="Errors to ignore (multiple)",
        )

        self.add_argument(
            "--exclude-file",
            dest="exclude_file",
            type=str,
            default=self._def_exclude_file,
            help="YAML file with errors to ignore",
        )

        # Ensure we initialize a namespace if needed,
        # and have a reference to it
        namespace = kwargs.get("namespace") or argparse.Namespace()
        kwargs["namespace"] = namespace
        # now parse args and put them in the namespace
        super().parse_args(*args, **kwargs)

        if namespace.exclude is None:
            namespace.exclude = []
        if os.path.exists(namespace.exclude_file):
            with open(namespace.exclude_file) as f:
                for i in yaml.safe_load(f):
                    if not self._expired(i):
                        namespace.exclude.append(i["error"])

        return namespace
