#!/usr/bin/env python3
# ------------------------------------------------
# This file is juju managed
# -----------------------------------------------
"""Process hplog output from cron and save for nagios check.

Copyright (C) 2008-2013  James Troup <james.troup@canonical.com>
Copyright (c) 2020 Rewrite: Drew Freiberger <drew.freiberger@canonical.com>

hplog lives in either /usr/sbin or /sbin depending on version of the
hp-health package and cron jobs don't get a sane $PATH
"""

import os
import subprocess
import sys

try:
    # shared lib will be under $NAGIOS_PLUGIN_DIR at runtime
    from hw_health_lib import (
        read_ignore_file,
        write_ignore_file,
        ignore,
        HPArgumentParser,
    )
except ImportError:
    # shared lib will be under $CHARM_SOURCE_DIR/files/common during unit tests
    common_libs_dir = os.path.abspath(
        os.path.join(os.path.dirname(__file__), "..", "common")
    )
    if common_libs_dir not in sys.path:
        sys.path.append(common_libs_dir)
    from hw_health_lib import (
        read_ignore_file,
        write_ignore_file,
        ignore,
        HPArgumentParser,
    )

FLAG_PROCESSOR = {
    "t": "parse_temperature",
    "f": "parse_fans",
    "p": "parse_power",
    "v": "save_log",
}
OUTPUT_FILE = "/var/lib/nagios/hplog.out"
EXCLUDE_FILE = "/etc/nagios/hplog.exclude.yaml"
DEBUG_FILE = "/var/lib/nagios/hplog.debug"


def parse_args(argv=None):
    """Define CLI arguments."""
    parser = HPArgumentParser(
        prog="cron_hplog", def_write_file=OUTPUT_FILE, def_exclude_file=EXCLUDE_FILE
    )
    parser.add_argument(
        "-f",
        "--hplog_flags",
        dest="hplog_flags",
        type=str,
        default=",".join(FLAG_PROCESSOR.keys()),
        help="Flags to call hplog with",
    )
    parser.add_argument(
        "-l",
        "--log_path",
        dest="log_path",
        type=str,
        default=DEBUG_FILE,
        help="Where to write hplog -v output for troubleshooting",
    )
    parser.add_argument(
        "-s",
        "--single_psu",
        dest="single_psu",
        action="store_true",
        help="Do not alert on lack of PSU redundancy",
        default=False,
    )
    return parser.parse_args(args=argv)


def call_hplog(flag):
    """Grab hplogs from hpasm if available.

    The output of this cron script will be checked by
    nagios via check_hplog.py.  The
    """
    env = "export PATH=$PATH:/usr/sbin:/sbin"

    cmd = "hplog -{}".format(flag)
    try:
        cmdline = "{}; {}".format(env, cmd)
        output = subprocess.check_output(cmdline, shell=True).decode("UTF-8")
    except subprocess.CalledProcessError as e:
        return 'Failed running command "{}" Return Code {}: {}' "".format(
            cmd, e.returncode, e.output
        )

    funcname = FLAG_PROCESSOR[flag]
    return globals()[funcname](output)


def parse_temperature(result):
    """Process the temperature results from hplog.

    Temperature [hplog -t]
              10        20        30        40        50        60        70
    01234567890123456789012345678901234567890123456789012345678901234567890
    ID     TYPE        LOCATION      STATUS    CURRENT  THRESHOLD
     1  Basic Sensor I/O Zone        Normal   102F/ 39C 158F/ 70C
     2  Basic Sensor Ambient         Normal    73F/ 23C 102F/ 39C
     3  Basic Sensor CPU (1)         Normal    95F/ 35C 260F/127C
     4  Basic Sensor CPU (1)         Normal    95F/ 35C 260F/127C
     5  Basic Sensor Pwr. Supply Bay Normal   111F/ 44C 170F/ 77C
     6  Basic Sensor CPU (2)         Normal   ---F/---C 260F/127C
     7  Basic Sensor CPU (2)         Normal   ---F/---C 260F/127C
    """
    input_file = result.splitlines()
    if os.path.isfile("/etc/nagios/skip-cat-hp-temperature.txt"):
        return
    header_line = input_file.pop(0).strip()
    if (
        header_line != "ID     TYPE        LOCATION      STATUS    CURRENT  THRESHOLD"
    ):  # noqa E501
        return "UNKNOWN Unrecognised header line in 'hplog -t' output"
    for line in input_file:
        line = line.rstrip()
        if not line:
            continue
        temp_id = line[0:3].strip()
        # temp_type = line[4:16].strip()
        temp_location = line[17:32].strip()
        temp_location = "%s (%s)" % (temp_location, temp_id)
        temp_status = line[33:41].strip()
        temp_current = line[42:51].split("/")[1].strip()
        temp_threshold = line[52:].split("/")[1].strip()
        if temp_status not in ["Normal", "Nominal", "Absent"]:
            return "%s: temperature is '%s' (%s / %s)" % (
                temp_location,
                temp_status,
                temp_current,
                temp_threshold,
            )

    return


def parse_fans(result):
    """Process the fan status from hplog.

    Fans [hplog -f]
              10        20        30        40        50        60        70
    01234567890123456789012345678901234567890123456789012345678901234567890
    ID     TYPE        LOCATION      STATUS  REDUNDANT FAN SPEED
     1  Var. Speed   I/O Zone        Normal     Yes     Medium ( 45)
     2  Var. Speed   I/O Zone        Normal     Yes     Medium ( 45)
     3  Var. Speed   Processor Zone  Normal     Yes     Medium ( 41)
     4  Var. Speed   Processor Zone  Normal     Yes     Low    ( 36)
     5  Var. Speed   Processor Zone  Normal     Yes     Low    ( 36)
     6  Var. Speed   Processor Zone  Normal     Yes     Low    ( 36)
    """
    input_file = result.splitlines()
    if os.path.isfile("/etc/nagios/skip-cat-hp-fans.txt"):
        return
    header_line = input_file.pop(0).strip()
    if header_line != "ID     TYPE        LOCATION      STATUS  REDUNDANT FAN SPEED":
        return "UNKNOWN Unrecognised header line in 'hplog -f' output"

    ignore_file = "/etc/nagios/ignores/ignores-cat-hp-fans.txt"
    ignores = read_ignore_file(ignore_file)
    for line in input_file:
        line = line.rstrip()
        if not line:
            continue
        fan_id = line[0:3].strip()
        fan_type = line[4:16].strip()
        fan_location = line[17:32].strip()
        fan_location = "%s (%s)" % (fan_location, fan_id)
        fan_status = line[33:40].strip()
        fan_redundant = line[41:50].strip()
        fan_speed = line[51:].strip()

        (return_now, msg) = process_fan_line(
            fan_type, fan_location, fan_status, fan_speed, fan_redundant, ignores
        )
        if return_now:
            return msg

    write_ignore_file(ignores, ignore_file)


def process_fan_line(
    fan_type, fan_location, fan_status, fan_speed, fan_redundant, ignores
):
    """Parse each line of fan output and check against known alertable values."""
    if fan_type == "Basic Fan":
        return (False, None)

    if fan_type not in ["Var. Speed", "Pwr. Supply", "Auto. Speed"]:
        return (
            True,
            "UNKNOWN %s: Unrecognised fan type '%s'" % (fan_location, fan_type),
        )

    if fan_status not in ["Normal", "Nominal"]:
        err = "%s: fans are '%s' (%s / Redundant: %s)" % (
            fan_location,
            fan_status,
            fan_speed,
            fan_redundant,
        )
        if not ignore(err, ignores):
            return (True, err)

    if fan_redundant not in ["Yes", "N/A"] and fan_type == "Var. Speed":
        err = "%s: fans are not redundant (%s / Status: %s)" % (
            fan_location,
            fan_speed,
            fan_redundant,
        )
        if not ignore(err, ignores):
            return (True, err)

    return (False, None)


def parse_power(result):
    """Process the power supply status from hplog.

    Power supplies [hplog -p]
              10        20        30        40        50        60        70
    01234567890123456789012345678901234567890123456789012345678901234567890
    ID     TYPE        LOCATION      STATUS  REDUNDANT
     1  Standard     Pwr. Supply Bay Normal     Yes
     2  Standard     Pwr. Supply Bay Normal     Yes
    """
    input_file = result.splitlines()
    if os.path.isfile("/etc/nagios/skip-cat-hp-power.txt"):
        return
    header_line = input_file.pop(0).strip()
    if header_line != "ID     TYPE        LOCATION      STATUS  REDUNDANT":
        return "UNKNOWN Unrecognised header line in 'hplog -p' output"

    ignore_file = "/etc/nagios/ignores/ignores-cat-hp-power.txt"
    ignores = read_ignore_file(ignore_file)

    for line in input_file:
        line = line.rstrip()
        if not line:
            continue
        power_id = line[0:3].strip()
        power_type = line[4:16].strip()
        power_location = line[17:32].strip()
        power_location = "%s (%s)" % (power_location, power_id)
        power_status = line[33:40].strip()
        # power_redundant = line[41:50].strip()
        if power_type != "Standard":
            err = "%s: Unrecognised power type '%s'" % (power_location, power_type)
            if not ignore(err, ignores):
                return "UNKNOWN {}".format(err)
        if not ARGS.single_psu and power_status not in ["Normal", "Nominal"]:
            err = "%s: power supply is '%s'" % (power_location, power_status)
            if not ignore(err, ignores):
                return err
    write_ignore_file(ignores, ignore_file)


def save_log(result):
    """Save full hplog -v output for troubleshooting after alert."""
    with open(ARGS.log_path, "w") as f:
        f.write(result)
    return


def main():
    """Define main subroutine to run in cron."""
    global ARGS
    ARGS = parse_args()

    try:
        # This matches hpasmlited on latest packages for bionic on <= gen9
        subprocess.check_call("ps -ef | grep -q hp[a]sm", shell=True)
    except subprocess.CalledProcessError as e:
        msg = (
            "UNKNOWN hp[a]sm daemon not found running, cannot run hplog: "
            "{}".format(e.output)
        )
        exit = 3
    else:
        errors = []
        for flag in ARGS.hplog_flags.split(","):
            log_output = call_hplog(flag)
            if log_output:
                errors.append(log_output)

        if len(errors) > 0:
            msg = "CRIT {} error(s): {}".format(len(errors), " - ".join(errors))
            exit = 2
        else:
            msg = "OK No errors found"
            exit = 0

    if ARGS.write:
        with open(ARGS.write, "w") as f:
            f.write(msg)
    else:
        print(msg)
    sys.exit(exit)


if __name__ == "__main__":
    main()
