#!/usr/bin/env python3
# ------------------------------------------------
# This file is juju managed
# ------------------------------------------------
"""ilorest interface for Gen10 HPE machines.

Original author: Ryan Finnie <ryan.finnie@canonical.com>
Porting for the HW health charm: Andrea Ieri <andrea.ieri@canonical.com>
"""

import json
import os
import sys
from subprocess import check_output

try:
    # shared lib will be under $NAGIOS_PLUGIN_DIR at runtime
    from hw_health_lib import HPArgumentParser
except ImportError:
    # shared lib will be under $CHARM_SOURCE_DIR/files/common during unit tests
    common_libs_dir = os.path.abspath(
        os.path.join(os.path.dirname(__file__), "..", "common")
    )
    if common_libs_dir not in sys.path:
        sys.path.append(common_libs_dir)
    from hw_health_lib import HPArgumentParser


DEFAULT_SELECTORS = [
    "Chassis",
    "HpeSmartStorage",
    "Memory",
    "Power",
    "Processor",
    "Thermal",
]

EXCLUDE_FILE = "/etc/nagios/cron_ilorest.exclude.yaml"
OUTPUT_FILE = "/var/lib/nagios/ilorest.out"


class CronILOrest:
    """Implement ilorest checks."""

    def __init__(self, argv=[]):
        """Parse CLI args during instance creation."""
        self.args = self.parse_args(argv)

    def parse_args(self, argv=None):
        """Define CLI arguments for cron_ilorest script."""
        parser = HPArgumentParser(
            prog="cron_ilorest",
            description=(
                "Convert the output of ilorest into an appropriate "
                "Nagios status line"
            ),
            def_write_file=OUTPUT_FILE,
            def_exclude_file=EXCLUDE_FILE,
        )

        parser.add_argument(
            "--selectors",
            dest="selectors",
            type=str,
            nargs="+",
            default=DEFAULT_SELECTORS,
            help="iLO selectors to run",
        )

        return parser.parse_args(args=argv)

    def check_selector(self, selector):
        """Check the selector status via ilorest.

        There are several interfaces called "selectors" in the ILO's restful interface.
        This queries the requested selector and checks it for known alertable states.
        """
        if self.args.debug:
            print("Checking selector {}".format(selector), file=sys.stderr)
        ilorest_output = self._get_json_ilorest_output(selector).splitlines()

        # Disregard the first few lines, json data starts at either { or [
        line = ""
        while line not in ["{", "["]:
            line = ilorest_output.pop(0)

        ilorest_output.insert(0, line)
        jsondata = json.loads("\n".join(ilorest_output))

        return self._walk_selector(jsondata, [selector])

    def _get_json_ilorest_output(self, selector):
        cmd = ["/usr/sbin/ilorest", "list", "-j", "--selector={}".format(selector)]
        return check_output(cmd).decode("UTF-8")

    def _get_health_status_message(self, json_item, crumb_trail=[]):
        desc = json_item["Name"]
        if "SerialNumber" in json_item:
            desc += " ({})".format(json_item["SerialNumber"])
        state = json_item.get("Status", "null").get("State", "unknown")
        health = json_item.get("Status", "null").get("Health", "unknown")
        msg = "{} ({}): {} health {}".format(" ".join(crumb_trail), desc, state, health)
        if self.args.debug:
            print(msg, file=sys.stderr)

        if msg in self.args.exclude and self.args.debug:
            print("Ignoring excluded error: {}".format(msg), file=sys.stderr)
            return []
        else:
            return [msg]

    def _walk_selector(self, json_data, crumb_trail=[]):
        errors = []

        # if we start at a list, iterate through it, else make sure we've got a dict
        if type(json_data) == list:
            for index in range(len(json_data)):
                errors.extend(
                    self._walk_selector(json_data[index], (crumb_trail + [str(index)]))
                )
        elif type(json_data) != dict:
            # we only dive through lists and dicts, so no errors below this point
            return []
        elif (
            json_data.get("Status")
            and json_data.get("Status").get("State") != "Absent"  # noqa:W503
            and json_data.get("Status").get("Health") != "OK"  # noqa:W503
        ):
            # we return the status message because there's no need to go deeper
            return self._get_health_status_message(json_data, crumb_trail)
        elif json_data.get("Status"):
            # found a status section locally but it's not alertable, eject
            return []
        else:
            # recurse through any other dicts or lists in the current node.
            for keyname in json_data.keys():
                if type(json_data[keyname]) in [list, dict]:
                    errors.extend(
                        self._walk_selector(
                            json_data[keyname], (crumb_trail + [keyname])
                        )
                    )

        return errors


def main(argv=None):
    """Define main subroutine for cron_ilorest."""
    cronilorest = CronILOrest(argv)

    errors = []
    for selector in cronilorest.args.selectors:
        errors.extend(cronilorest.check_selector(selector))

    if len(errors) > 0:
        msg = "CRIT {} error(s): {}".format(len(errors), " - ".join(errors))
        exit = 2
    else:
        msg = "OK No errors found"
        exit = 0

    if cronilorest.args.write:
        if cronilorest.args.write == "-":
            print(msg)
        else:
            with open(cronilorest.args.write, "w") as f:
                f.write(msg)
    else:
        # This should never happen since 'write' has a default value
        print(msg)
    sys.exit(exit)


if __name__ == "__main__":
    main(sys.argv[1:])
