#!/usr/bin/env python3
"""This script checks ipmi sensors for known alertable states."""

import os
import subprocess
import sys

CHECK_IPMI_PID = "/var/run/nagios/check_ipmi_sensors.pid"
OUTPUT_FILE = "/var/lib/nagios/ipmi_sensors.out"
TMP_OUTPUT_FILE = OUTPUT_FILE + ".tmp"
CMD = "/usr/local/lib/nagios/plugins/check_ipmi_sensor"
NAGIOS_ERRORS = {
    1: "WARNING",
    2: "CRITICAL",
    3: "UNKNOWN",
}


def write_output_file(output):
    """Write results of checks to the defined location for nagios to check."""
    try:
        with open(TMP_OUTPUT_FILE, "w") as fd:
            fd.write(output)
    except IOError as e:
        print("Cannot write output file {}, error {}".format(TMP_OUTPUT_FILE, e))
        sys.exit(1)
    os.rename(TMP_OUTPUT_FILE, OUTPUT_FILE)


def gather_metrics():
    """Execute ipmi tools to gather sensor metrics."""
    # Check if a PID file exists
    if os.path.exists(CHECK_IPMI_PID):
        # is the PID valid?
        with open(CHECK_IPMI_PID, "r") as fd:
            pid = fd.read()
        if pid not in os.listdir("/proc"):
            # PID file is invalid, remove it
            os.remove(CHECK_IPMI_PID)
        else:
            return

    try:
        with open(CHECK_IPMI_PID, "w") as fd:
            fd.write(str(os.getpid()))
    except IOError as e:
        # unable to write PID file, can't lock
        print("Cannot write lock file {}, error {}".format(CHECK_IPMI_PID, e))
        sys.exit(1)

    cmdline = [CMD]
    if len(sys.argv) > 1:
        cmdline.extend(sys.argv[1:])
    try:
        output = subprocess.check_output(cmdline).decode("utf8")
        write_output_file(output)
    except subprocess.CalledProcessError as error:
        output = error.stdout.decode(errors="ignore")
        write_output_file("{}: {}".format(NAGIOS_ERRORS[error.returncode], output))
    except PermissionError as error:
        write_output_file("UNKNOWN: {}".format(error))

    # remove pid reference
    os.remove(CHECK_IPMI_PID)


if __name__ == "__main__":
    gather_metrics()
