#!/usr/bin/env python3
# -*- coding: us-ascii -*-
"""Cronjob to process mdadm software raid status for known bad states."""

import argparse
import os
import re
import shutil
import subprocess
import sys

try:
    # shared lib will be under $NAGIOS_PLUGIN_DIR at runtime
    from hw_health_lib import HWCronArgumentParser
except ImportError:
    # shared lib will be under $CHARM_SOURCE_DIR/files/common during unit tests
    common_libs_dir = os.path.abspath(
        os.path.join(os.path.dirname(__file__), "..", "common")
    )
    if common_libs_dir not in sys.path:
        sys.path.append(common_libs_dir)
    from hw_health_lib import HWCronArgumentParser

OUTPUT_FILE = "/var/lib/nagios/mdadm.out"
TEMP_FILE = "/tmp/mdadm.out"
ARGS = argparse.Namespace()


def get_devices():
    """Get a list of software raid devices to check."""
    if os.path.exists("/sbin/mdadm"):
        try:
            cmd = ["/sbin/mdadm", "--detail", "--scan"]
            devices_raw = subprocess.check_output(cmd)
            devices_re = re.compile(r"^ARRAY\s+([^ ]+) ")
            devices = set()
            for line in devices_raw.decode().split("\n"):
                line = line.strip()
                device_re = devices_re.match(line)
                if device_re is not None:
                    devices.add(device_re.group(1))
            return devices
        except subprocess.CalledProcessError as error:
            rc = generate_output("CRITICAL: get_devices error - {}".format(error))
            if rc:
                sys.exit(0)
    return set()


def generate_output(msg):
    """Write output file for nagios check to read results from."""
    try:
        with open(TEMP_FILE, "w") as fd:
            fd.write(msg)
        shutil.move(TEMP_FILE, ARGS.write)
        return True
    except Exception as error:
        print("Unable to generate output file:", error)
        return False


def get_devices_stats(devices):
    """Check software raid device for known alertable states."""
    mdadm_detail = ["/sbin/mdadm", "--detail"]
    mdadm_detail.extend(sorted(devices))

    devices_details_raw = subprocess.check_output(mdadm_detail)

    devices_re = r"^(/\S+):$"
    state_re = r"^\s*State\s+:\s+(.+)\s*$"
    status_re = r"^\s*(Active|Working|Failed|Spare) Devices\s+:\s+(\d+)$"
    rebuild_status_re = r"^\s*Rebuild Status\s+:\s+(\d+%\s+\S+)$"
    removed_re = r"^\s*-\s+0\s+0\s+(\d+)\s+removed$"
    #        4 8 162 3 spare rebuilding /dev/sdk2
    rebuilding_re = r"^\s*\d+\s+\d+\s+\d+\s+\d+\s+\S+\s+rebuilding\s+(\S+)$"

    devices_cre = re.compile(devices_re)
    state_cre = re.compile(state_re)
    status_cre = re.compile(status_re)
    rebuild_status_cre = re.compile(rebuild_status_re)
    removed_cre = re.compile(removed_re)
    rebuilding_cre = re.compile(rebuilding_re)

    device = None
    devices_stats = {}
    for line in devices_details_raw.decode().split("\n"):
        line = line.rstrip()
        m = devices_cre.match(line)
        if m:
            device = m.group(1)
            devices_stats[device] = {
                "stats": {"Active": 0, "Working": 0, "Failed": 0, "Spare": 0},
                "rebuild_status": "",
                "degraded": False,
                "recovering": False,
                "removed": [],
                "rebuilding": [],
            }
            continue

        m = state_cre.match(line)
        if m:
            # format for State line can be "clean" or "clean, degraded",
            # or "active, degraded, rebuilding", etc.
            states = m.group(1).split(", ")
            if "degraded" in states and device:
                devices_stats[device]["degraded"] = True
            if "recovering" in states and device:
                devices_stats[device]["recovering"] = True
            continue

        m = status_cre.match(line)
        if m and device:
            devices_stats[device]["stats"][m.group(1)] = int(m.group(2))
            continue

        m = removed_cre.match(line)
        if m and device:
            devices_stats[device]["removed"].append(m.group(1))
            continue

        m = rebuild_status_cre.match(line)
        if m and device:
            devices_stats[device]["rebuild_status"] = m.group(1)
            continue

        m = rebuilding_cre.match(line)
        if m and device:
            devices_stats[device]["rebuilding"].append(m.group(1))
            continue

    return devices_stats


def parse_output():  # noqa:C901
    devices = get_devices()
    if len(devices) == 0:
        return generate_output("WARNING: unexpectedly checked no devices")

    try:
        devices_stats = get_devices_stats(devices)
    except subprocess.CalledProcessError as error:
        return generate_output("WARNING: error executing mdadm: {}".format(error))

    msg = []
    critical = False
    warning = False
    for device in devices_stats:
        parts = []
        # Is device degraded?
        if devices_stats[device]["degraded"] and devices_stats[device]["recovering"]:
            warning = True
            parts = ["{} recovering".format(device)]
        elif devices_stats[device]["degraded"]:
            critical = True
            parts = ["{} degraded".format(device)]
        else:
            parts = ["{} ok".format(device)]

        # If Failed drives are found, list counters (how many?)
        failed_cnt = devices_stats[device]["stats"].get("Failed", 0)
        if failed_cnt > 0:
            critical = True
            dev_stats = [
                "{}[{}]".format(status, devices_stats[device]["stats"][status])
                for status in sorted(devices_stats[device]["stats"])
            ]
            parts.extend(dev_stats)

        if len(devices_stats[device]["removed"]) != 0:
            critical = True
            members = " and ".join(devices_stats[device]["removed"])
            parts.append("RaidDevice(s) {} marked removed".format(members))

        if len(devices_stats[device]["rebuilding"]) != 0:
            rebuilding_members = " ".join(devices_stats[device]["rebuilding"])
            rebuild_status = devices_stats[device]["rebuild_status"]
            parts.append(
                "{} rebuilding ({})".format(rebuilding_members, rebuild_status)
            )

        msg.append(", ".join(parts))

    if critical:
        msg = "CRITICAL: {}".format("; ".join(msg))
    elif warning:
        msg = "WARNING: {}".format("; ".join(msg))
    else:
        msg = "OK: {}".format("; ".join(msg))
    return generate_output(msg)


def parse_args(argv=None):
    """Process CLI arguments."""
    parser = HWCronArgumentParser(prog="cron_mdadm", def_write_file=OUTPUT_FILE)
    return parser.parse_args(args=argv, namespace=ARGS)


def main(argv):
    """Define main subroutine."""
    parse_args(argv)
    parse_output()


if __name__ == "__main__":
    main(sys.argv[1:])
