#!/usr/bin/env python3
# -*- coding: us-ascii -*-
"""Check for issues with NVME hardware devices."""

import argparse
import glob
import re
import subprocess
import sys

from nagios_plugin3 import CriticalError, UnknownError, try_check

NVME_RE = re.compile(r"^/dev/nvme\d+$")


def parse_output():
    """Process nvme smart-log information and alert on known issues."""
    keymap = {}
    critical = False
    alloutputs = []
    for device in glob.glob("/dev/nvme*"):
        if not NVME_RE.match(device):
            continue
        try:
            output = subprocess.check_output(
                ["sudo", "/usr/sbin/nvme", "smart-log", device]
            )
        except subprocess.CalledProcessError as error:
            print("nvme check error: {}".format(error))
            return

        for line in output.decode(errors="ignore").splitlines():
            datavalues_re = re.match(r"^(\w+)\s+:\s+([\d.]+)", line.strip())
            if not datavalues_re:
                continue
            key, value = datavalues_re.groups()
            keymap[key] = value.replace(".", "")

        if int(keymap["critical_warning"]) != 0:
            status = ("CRITICAL: {} critical_warning is {}").format(
                device, keymap["critical_warning"]
            )
            critical = True
        else:
            status = "OK: no errors on {}".format(device)

        alloutputs.append(
            "{} | {}".format(
                status,
                " ".join(
                    ["{}={}".format(repr(key), value) for key, value in keymap.items()]
                ),
            )
        )

    if critical:
        raise CriticalError("\n".join(alloutputs))

    if not alloutputs:
        raise UnknownError("no nvme devices found")

    print("\n".join(alloutputs))


def parse_args(argv=None):
    """Process CLI arguments."""
    parser = argparse.ArgumentParser(
        prog="check_nvme",
        description=(
            "this program reads the nvme smart-log and outputs an "
            "appropriate Nagios status line"
        ),
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    return parser.parse_args(argv)


def main(argv):
    """Define main subroutine."""
    parse_args(argv)
    try_check(parse_output)


if __name__ == "__main__":
    main(sys.argv[1:])
