#!/usr/bin/env python3
# -*- coding: us-ascii -*-
"""Check sas2ircu output for known alertable states."""

import argparse
import os
import re
import sys

from nagios_plugin3 import CriticalError, WarnError, try_check

try:
    # shared lib will be under $NAGIOS_PLUGIN_DIR at runtime
    from hw_health_lib import HWCheckArgumentParser
except ImportError:
    # shared lib will be under $CHARM_SOURCE_DIR/files/common during unit tests
    common_libs_dir = os.path.abspath(
        os.path.join(os.path.dirname(__file__), "..", "common")
    )
    if common_libs_dir not in sys.path:
        sys.path.append(common_libs_dir)
    from hw_health_lib import HWCheckArgumentParser

INPUT_FILE = "/var/lib/nagios/sas2ircu.out"
ARGS = argparse.Namespace()


def parse_output():
    """Parse sas2ircu output and raise known alertable states."""
    enclosure_re = r"^\s+Enclosure #\s+:\s+(\d+)"
    slot_re = r"^\s+Slot #\s+:\s+(\d+)"
    state_re = r"^\s+State\s+:\s+(\S+)"

    encl_slot_state_cre = [
        re.compile(enclosure_re),
        re.compile(slot_re),
        re.compile(state_re),
    ]

    devices = {}
    device = []
    critical = False
    with open(ARGS.input_file) as devices_raw:
        for line in devices_raw.readlines():
            line = line.rstrip()
            for cre in encl_slot_state_cre:
                m = cre.match(line)
                if m:
                    device.append(m.group(1))

            if len(device) == 3:
                tmpdev = devices.get(device[2], [])
                tmpdev.append("{}:{}".format(device[0], device[1]))
                devices[device[2]] = tmpdev
                if not ("Ready" in device or "Optimal" in device):
                    critical = True
                device = []

    msg = "; ".join(
        ["{}[{}]".format(state, ",".join(devices[state])) for state in devices]
    )
    if msg == "":
        raise WarnError("WARNING: no output")
    elif critical:
        raise CriticalError("CRITICAL: {}".format(msg))
    else:
        print("OK: {}".format(msg))


def parse_args(argv=None):
    """Process CLI arguments."""
    parser = HWCheckArgumentParser(prog="check_sas2ircu", def_input_file=INPUT_FILE)
    return parser.parse_args(args=argv, namespace=ARGS)


def main(argv):
    """Define main subroutine."""
    parse_args(argv)
    try_check(parse_output)


if __name__ == "__main__":
    main(sys.argv[1:])
