#!/usr/bin/env python3
# -*- coding: us-ascii -*-
"""This script implements a nagios check for output from the vendor tool "sas3ircu"."""

import argparse
import os
import re
import sys
from collections import defaultdict

from nagios_plugin3 import CriticalError, UnknownError, WarnError, try_check

try:
    # shared lib will be under $NAGIOS_PLUGIN_DIR at runtime
    from hw_health_lib import HWCheckArgumentParser
except ImportError:
    # shared lib will be under $CHARM_SOURCE_DIR/files/common during unit tests
    common_libs_dir = os.path.abspath(
        os.path.join(os.path.dirname(__file__), "..", "common")
    )
    if common_libs_dir not in sys.path:
        sys.path.append(common_libs_dir)
    from hw_health_lib import HWCheckArgumentParser

INPUT_FILE = "/var/lib/nagios/sas3ircu.out"
ARGS = argparse.Namespace()


def parse_output(input_file):
    """Turn the whole sas3ircu output into a dictionary."""
    sections_re = re.compile(
        r"(?<=^Controller information\n)"
        r"-+\n"
        r"(?P<ctrl>(?:.|\n)*)"
        r"^-+\n"
        r"^IR Volume information\n"
        r"-+\n"
        r"(?P<vols>(?:.|\n)*)"
        r"^-+\n"
        r"^Physical device information\n"
        r"-+\n"
        r"(?P<disks>(?:.|\n)*)"
        r"^-+\n"
        r"^Enclosure information\n"
        r"-+\n"
        r"(?P<encl>(?:.|\n)*)"
        r"^-+\n",
        re.MULTILINE,
    )
    disks_re = re.compile(
        r"(?<=^Device is a Hard disk\n)(?P<kv_data>(?:.|\n)*?)(?=^$)", re.MULTILINE
    )

    with open(input_file) as devices_raw:
        sections = sections_re.search(devices_raw.read()).groupdict()
        controller = _kv_parse(sections["ctrl"])
        volumes = _vols_parse(sections["vols"])
        # This collects disk level information in a structure simulating the
        # physical encl/slot arrangement
        topology = defaultdict(dict)
        for match in disks_re.findall(sections["disks"]):
            disk = _kv_parse(match)
            encl = disk["Enclosure #"]
            slot = disk["Slot #"]
            topology[encl][slot] = disk
        enclosure = _kv_parse(sections["encl"])

    return {
        "controller": controller,
        "volumes": volumes,
        "disks": topology,
        "enclosure": enclosure,
    }


def _vols_parse(text):
    """Parse raw sas3ircu output and return volume data structure."""
    vols_re = re.compile(
        r"^IR volume (?P<n>\d+)\n"
        r"(?P<kv_data>(?:.|\n)*?)"
        r"\s+Physical hard disks\s+:.*\n"
        r"(?P<topology>(?:^\s+PHY.*\n)+)",
        re.MULTILINE,
    )
    vol_topology_re = re.compile(
        r"\s+PHY\[(?P<n>\d+)\]\s+Enclosure#\/Slot#\s+" r":\s+(?P<enc>\d+):(?P<slot>\d+)"
    )
    volumes = {}
    for (vol_n, kv_data, vol_topology) in vols_re.findall(text):
        topology = {}
        for (member_n, enc, slot) in vol_topology_re.findall(vol_topology):
            topology[member_n] = {"enc": enc, "slot": slot}
        volumes[vol_n] = {**_kv_parse(kv_data), "topology": topology}

    return volumes


def _kv_parse(text):
    """Build a dict by parsing " key : value" text data."""
    key_value_re = re.compile(r"^\s*(?P<key>.*?)\s+:\s+(?P<value>.*)")
    text = text.strip()
    return {
        m.group("key"): m.group("value")
        for m in map(key_value_re.search, text.split("\n"))
    }


def eval_status(data):
    """Determine the state of the storage subsystem.

    Input data is a dictionary and a set of rules
    """
    # Define known states
    ok = "Okay (OKY)"
    ready = "Ready (RDY)"
    optimal = "Optimal (OPT)"
    status = Status()

    # 1. Volumes must be in Okay state
    for volume in data["volumes"].values():
        vol_id = volume["Volume ID"]
        vol_status = volume["Status of volume"]
        if vol_status != ok:
            status.crit("Volume {}: {}".format(vol_id, vol_status))
        else:
            # 2. Volume members must be in Optimal state
            for member in volume["topology"].values():
                disk = data["disks"][member["enc"]][member["slot"]]
                if disk["State"] != optimal:
                    msg = "Disk {}:{} {}".format(
                        member["enc"], member["slot"], disk["State"]
                    )
                    if disk["State"] == ready:
                        status.warn(msg)
                    else:
                        status.crit(msg)
    # 3. Disks can be in Optimal or Ready state ("ready" is ok for non-RAID
    # members)
    for enclosure_id, enclosure in data["disks"].items():
        for slot_id, slot in enclosure.items():
            if slot["State"] not in [optimal, ready]:
                status.crit(
                    "Disk {}:{} {}".format(enclosure_id, slot_id, slot["State"])
                )
    status.get_status()


class Status:
    """Class hiding the whole "CRIT >> WARN >> OK" priority scheme."""

    def __init__(self, status="OK"):
        """Start status as blank and OK."""
        self._status = status
        self._msgs = set()

    def crit(self, msg):
        """Set critical status with meaningful message."""
        self._status = "CRITICAL"
        self._msgs.add(msg)

    def warn(self, msg):
        """Set warning status with meaningful message, unless already critical."""
        if self._status != "CRITICAL":
            self._status = "WARNING"
        self._msgs.add(msg)

    def ok(self, msg):
        """Add info messages to output without changing status from OK."""
        self._msgs.add(msg)

    def get_status(self):
        """Render the current status.

        Raises nagios_plugin3 exceptions if things are not OK
        """
        if self._status == "OK":
            msg = "{}: no errors".format(self._status)
            print(msg)
        else:
            msg = "{}: {}".format(self._status, " | ".join(self._msgs))
            if self._status == "CRITICAL":
                raise CriticalError(msg)
            elif self._status == "WARNING":
                raise WarnError(msg)
            else:
                # this really shouldn't be happening
                raise UnknownError(msg)

    def __repr__(self):
        """Define the official string-representation of the object."""
        print(self._status)


def parse_args(argv=None):
    """Process command line arguments and return an argparse object."""
    parser = HWCheckArgumentParser(prog="check_sas3ircu", def_input_file=INPUT_FILE)
    return parser.parse_args(args=argv, namespace=ARGS)


def main(argv=None):
    """Define main subroutine.

    Read cli args
    Parse output of the cron-collected file
    Return nagios-friendly response codes and info.
    """
    parse_args(argv)
    data = parse_output(ARGS.input_file)
    try_check(eval_status, data)


if __name__ == "__main__":
    main(sys.argv[1:])
