#!/usr/bin/env python3
# -----------------------------------------------
# This file is juju managed
# -----------------------------------------------

# Copyright (C) 2008-2013  James Troup <james.troup@canonical.com>
# Copyright (c) 2020 Rewrite: Drew Freiberger <drew.freiberger@canonical.com>

# ssacli lives in either /usr/sbin or /sbin depending on version of the
# package and cron jobs don't get a sane $PATH
"""Capture ssacli output for future nagios check processing."""

import os
import re
import subprocess
import sys

try:
    # shared lib will be under $NAGIOS_PLUGIN_DIR at runtime
    from hw_health_lib import (
        read_ignore_file,
        write_ignore_file,
        ignore,
        get_hp_controller_slots,
        HPArgumentParser,
        SSACLI_BIN,
    )
except ImportError:
    # shared lib will be under $CHARM_SOURCE_DIR/files/common during unit tests
    common_libs_dir = os.path.abspath(
        os.path.join(os.path.dirname(__file__), "..", "common")
    )
    if common_libs_dir not in sys.path:
        sys.path.append(common_libs_dir)
    from hw_health_lib import (
        read_ignore_file,
        write_ignore_file,
        ignore,
        get_hp_controller_slots,
        HPArgumentParser,
        SSACLI_BIN,
    )

OUTPUT_FILE = "/var/lib/nagios/ssacli.out"
EXCLUDE_FILE = "/etc/nagios/ssacli.exclude.yaml"


def parse_args(argv=None):
    """Process CLI arguments."""
    parser = HPArgumentParser(
        prog="cron_ssacli", def_write_file=OUTPUT_FILE, def_exclude_file=EXCLUDE_FILE
    )
    return parser.parse_args(args=argv)


def check_array(slot):
    """Process ssacli output for alertable statuses on logical and physical disks.

    RAID arrays [hpacucli]

    logicaldrive 1 (68.3 GB, RAID RAID 1+0):  OK
    logicaldrive 1 (2.8 TB, RAID RAID 6 (ADG)):  OK
    physicaldrive 1:1 (box 1:bay 1, 72 GB): OK
    physicaldrive 1:2 (box 1:bay 2, 72 GB): OK
    """
    if os.path.isfile("/etc/nagios/skip-cat-hp-array.txt"):
        return

    cmd = (
        "{ssacli} ctrl slot={slot} ld all show status; "
        "{ssacli} ctrl slot={slot} pd all show status".format(
            ssacli=SSACLI_BIN, slot=slot
        )
    )
    try:
        result = subprocess.check_output(cmd, shell=True).decode("UTF-8")
        return _parse_array_output(result)
    except subprocess.CalledProcessError as e:
        return (
            "UNKNOWN Call to ssacli to show ld/pd info failed. "
            "Array Slot {} - Return Code {} - {}"
            "".format(slot, e.returncode, e.output)
        )


def _parse_array_output(output):
    innocuous_errors = re.compile(
        r"^Error: The specified (device|controller) "
        "does not have any (logical|physical)"
    )
    drive_status_line = re.compile(r"^\s*(logicaldrive|physicaldrive)")
    ignore_file = "/etc/nagios/ignores/ignores-cat-hp-array.txt"
    ignores = read_ignore_file(ignore_file)

    for line in output.splitlines():
        line = line.strip()
        if (
            not line
            or innocuous_errors.search(line)  # noqa:W503
            or not drive_status_line.search(line)  # noqa:W503
        ):
            continue
        (drivetype, number) = line.split()[:2]
        status = line.split("):")[1].lstrip().upper()
        if status != "OK":
            err = '{} {} is "{}"'.format(drivetype, number, status)
            if not ignore(err, ignores):
                return err
    write_ignore_file(ignores, ignore_file)


def check_controller(slot):
    """Check for known bad states of controller.

    RAID controllers [hpacucli]

    Smart Array 6400 in Slot 1
      Controller Status: OK
      Cache Status: OK
      Battery Status: Failed (Replace Batteries)
    """
    if os.path.isfile("/etc/nagios/skip-cat-hp-controller.txt"):
        return

    cmd = "{ssacli} ctrl slot={slot} show status".format(ssacli=SSACLI_BIN, slot=slot)
    try:
        result = subprocess.check_output(cmd, shell=True).decode("UTF-8")
        return _parse_controller_output(result)
    except subprocess.CalledProcessError as e:
        return (
            "UNKNOWN Call to ssacli to show ld/pd info failed. "
            "Array Slot {} - Return Code {} - {}"
            "".format(slot, e.returncode, e.output)
        )


def _parse_controller_output(output):
    controller = "Unknown"
    ignore_file = "/etc/nagios/ignores/ignores-cat-hp-controller.txt"
    ignores = read_ignore_file(ignore_file)
    for line in output.splitlines():
        line = line.strip()
        if not line:
            continue
        if line.startswith("Smart Array") or line.startswith("Smart HBA"):
            controller = line
        else:
            if ":" not in line:
                err = "UNKNOWN Unrecognised line for controller '%s'" % (line)
                return err
            (part, status) = line.split(":")
            status = status.strip().upper()
            if status != "OK":
                err = "%s on %s is '%s'" % (part, controller, status)
                if not ignore(err, ignores):
                    return err
    write_ignore_file(ignores, ignore_file)


def main():
    """Define main subroutine.

    Process CLI args
    Check status of controllers
    Check status of array disks and logical volumes
    Raise nagios exceptions as appropriate.
    """
    global ARGS
    ARGS = parse_args()

    slots = get_hp_controller_slots()
    if not slots:
        msg = "OK: no controller/array found to check"
        exit = 0

    errors = []
    for slot in slots:
        errors.append(check_controller(slot))
        errors.append(check_array(slot))
    errors = [e for e in errors if e]

    if len(errors) > 0:
        msg = "CRIT {} error(s): {}".format(len(errors), " - ".join(errors))
        exit = 2
    else:
        msg = "OK No errors found"
        exit = 0

    if ARGS.write:
        with open(ARGS.write, "w") as f:
            f.write(msg)
    else:
        print(msg)
    sys.exit(exit)


if __name__ == "__main__":
    main()
