# -*- coding: us-ascii -*-
"""Discover hardware and determine tools selection."""
import os
import re
import subprocess
from itertools import chain

from charmhelpers.core import hookenv

from hwhealth import tools
from hwhealth.discovery.lshw import Hardware
from hwhealth.discovery.supported_vendors import (
    SUPPORTED_DRIVERS,
    SUPPORTED_STORAGE,
    SUPPORTED_SYSTEMS,
)


def get_tools(manufacturer="auto"):
    """Return list of tool classes relevent for the current hardware.

    In testing, we set manufacturer = test in order to test all tools classes.
    Filtering added for known bad tools that don't work on all series combinations.
    """
    if manufacturer == "test":
        # Return all possible tools to aid testing
        storage_tools = {
            tool for vendor in SUPPORTED_STORAGE.values() for tool in vendor.values()
        }
        # Some system vendors have multiple tools, have to iterate sets
        system_tools = set(chain.from_iterable(SUPPORTED_SYSTEMS.values()))
        driver_tools = set(SUPPORTED_DRIVERS.values())
        ipmi_tools = set()
        if hookenv.config("enable_ipmi"):
            ipmi_tools = {tools.Ipmi}
        all_tools = storage_tools | system_tools | driver_tools | ipmi_tools
        series_filtered_tools = set(
            tool for tool in all_tools if tool.is_series_supported()
        )
        return series_filtered_tools
    elif manufacturer == "auto":
        return _get_tools()
    else:
        raise NotImplementedError


def _get_tools_os(toolset):
    """Handle freeipmi and mdadm packages, included with Ubuntu."""

    if _supports_mdadm():
        toolset.add(tools.Mdadm)

    if hookenv.config("enable_ipmi"):
        toolset.add(tools.Ipmi)

    if hookenv.config("enable_ipmiseld"):
        toolset.add(tools.IpmiSEL)


def _get_tools():
    """Return list of tool classes relevent for the current hardware."""
    hwinfo = Hardware()
    toolset = set()
    for storage in hwinfo.get_storage_class_info:
        vendor = storage.get("vendor")
        product = storage.get("product")
        tool = SUPPORTED_STORAGE.get(vendor, {}).get(product)
        if isinstance(tool, list) or isinstance(tool, set):
            toolset.update(tool)
        elif tool:
            toolset.add(tool)
        else:
            hookenv.log(
                "Product not supported: [{}][{}]".format(vendor, product),
                hookenv.DEBUG,
            )

        driver = storage.get("driver")
        if driver:
            if driver in SUPPORTED_DRIVERS:
                toolset.add(SUPPORTED_DRIVERS[driver])
                continue
            hookenv.log("Driver not supported: {}".format(driver), hookenv.DEBUG)

    _get_tools_os(toolset)

    system_vendor = hwinfo.get_system.get("vendor")
    tool = SUPPORTED_SYSTEMS.get(system_vendor)
    if isinstance(tool, list) or isinstance(tool, set):
        toolset.update(tool)
    elif tool:
        toolset.add(tool)
    else:
        hookenv.log(
            "System vendor not supported: {}".format(system_vendor), hookenv.DEBUG
        )

    toolset = set(tool for tool in toolset if tool.is_series_supported)
    return toolset


def _supports_mdadm():
    """Detect known mdadm supported storage devices.

    Returns True when the first one is found; otherwise, it returns False)
    """
    if os.path.exists("/sbin/mdadm"):
        try:
            devices_raw = subprocess.check_output(["/sbin/mdadm", "--detail", "--scan"])
            devices_re = re.compile(r"^ARRAY\s+(\S+) ")
            for line in devices_raw.splitlines():
                line = line.decode().strip()
                raid_dev = devices_re.search(line)
                if raid_dev:
                    hookenv.log("Found md raid array {}".format(raid_dev.group(1)))
                    return True
        except Exception as e:
            hookenv.log("mdadm scan failed with {}".format(e))
    return False


if __name__ == "__main__":
    toolset = get_tools()
    if not toolset:
        print("No RAID")
    else:
        print(toolset)
