# -*- coding: us-ascii -*-

import binascii
import glob
import hashlib
import os
import re
import shlex
import shutil
import subprocess
from pathlib import Path
from tempfile import TemporaryDirectory
from zipfile import BadZipFile, ZipFile

from charmhelpers import fetch
from charmhelpers.contrib.charmsupport import nrpe
from charmhelpers.core import hookenv, unitdata
from charmhelpers.core.host import lsb_release, write_file
from charmhelpers.core.templating import render

from charms import apt

import yaml


class JujuResourceNotFound(Exception):
    """Raise due to resource needed but not attached."""

    pass


class ToolError(Exception):
    """Allow a dict to be shared with try-except."""

    def __init__(self, keymap):
        self._keymap = keymap

    @property
    def message(self):
        return self._keymap


class ToolChecksumError(ToolError):
    """Raise due to resource does not match a whitelisted checksum."""

    pass


class ToolNotFound(ToolError):
    """Raise if resource found (zipfile) but does not contain a needed binary."""

    pass


class Tool:
    """An abstract class representing a "tool".

    The idea is to delegate install/configure duties to specific per-tool
    classes instead of handling multiple exceptional cases in a main loop
    during the install hook.
    Every tool should implement its own internal logic regarding how to be
    installed, configured, and removed.
    """

    CROND_DIR = "/etc/cron.d"
    CRONJOB_SCRIPT_MODE = 0o100755
    CRONJOB_SCRIPT_UID = 0
    CRONJOB_SCRIPT_GID = 0
    CRONJOB_OUTPUT_DIR = "/var/lib/nagios"
    NRPE_PLUGINS_DIR = "/usr/local/lib/nagios/plugins"
    NRPE_PLUGINS_MODE = 0o100755
    NRPE_PLUGINS_UID = 0
    NRPE_PLUGINS_GID = 0
    SUDOERS_DIR = "/etc/sudoers.d"
    SUDOERS_MODE = 0o100440
    SUDOERS_UID = 0
    SUDOERS_GID = 0
    SUPPORTED_SERIES = ["xenial", "bionic", "focal"]
    TOOLS_DIR = "/usr/local/bin"
    TOOLS_MODE = 0o100755
    TOOLS_UID = 0
    TOOLS_GID = 0

    def __init__(
        self,
        shortname=None,
        nrpe_opts="",
        nrpe_script=None,
        nrpe_script_dir=None,
        cron_script=None,
        cron_script_dir=None,
        cron_script_args=None,
    ):
        self._nagios_hostname = nrpe.get_nagios_hostname()
        self._nrpe_opts = nrpe_opts
        self._shortname = shortname if shortname else self.__class__.__name__.lower()
        self._files_dir = os.path.join(hookenv.charm_dir(), "files", self._shortname)
        self._nrpe_script = (
            nrpe_script if nrpe_script else "check_{}.py".format(self._shortname)
        )
        self._nrpe_script_dir = nrpe_script_dir if nrpe_script_dir else self._files_dir
        self._cron_script = (
            cron_script if cron_script else "cron_{}.py".format(self._shortname)
        )
        self._cron_script_dir = cron_script_dir if cron_script_dir else self._files_dir
        self._templates_dir = os.path.join(
            hookenv.charm_dir(), "templates", self._shortname
        )
        self._common_libs_dir = os.path.join(hookenv.charm_dir(), "files/common")
        self._common_libs = ["hw_health_lib.py"]
        self._cron_script_args = cron_script_args

    def _install_nrpe_plugin(self):
        src = os.path.join(self._nrpe_script_dir, self._nrpe_script)
        dst = shutil.copy(src, self.NRPE_PLUGINS_DIR)
        os.chmod(dst, self.NRPE_PLUGINS_MODE)
        os.chown(dst, uid=self.NRPE_PLUGINS_UID, gid=self.NRPE_PLUGINS_GID)
        hookenv.log(
            "NRPE script for tool [{}] installed at as {}"
            "".format(self._shortname, dst),
            hookenv.DEBUG,
        )

        return dst

    def _install_common_libs(self):
        dsts = []

        for lib in self._common_libs:
            src = os.path.join(self._common_libs_dir, lib)
            dst = shutil.copy(src, self.NRPE_PLUGINS_DIR)
            os.chmod(dst, self.NRPE_PLUGINS_MODE)
            os.chown(dst, uid=self.NRPE_PLUGINS_UID, gid=self.NRPE_PLUGINS_GID)
            hookenv.log(
                "Common Library {} for tool [{}] installed at as {}"
                "".format(lib, self._shortname, dst),
                hookenv.DEBUG,
            )
            dsts.append(dst)

        return dsts

    def _remove_nrpe_plugin(self):
        plugin_path = os.path.join(self.NRPE_PLUGINS_DIR, self._nrpe_script)

        if not os.path.exists(plugin_path):
            return
        os.remove(plugin_path)
        hookenv.log(
            "deleted NRPE script for tool [{}]".format(self._shortname), hookenv.DEBUG
        )

    def configure_nrpe_check(self, nrpe_setup):
        cmd = " ".join([os.path.basename(self._nrpe_script), self._nrpe_opts])
        nrpe_setup.add_check(
            shortname=self._shortname,
            description="{} Hardware Health".format(self._shortname),
            check_cmd=cmd,
        )
        hookenv.log(
            "configured NRPE check for tool [{}]".format(self._shortname), hookenv.DEBUG
        )

    def remove_nrpe_check(self):
        nrpe_setup = nrpe.NRPE(hostname=self._nagios_hostname, primary=False)
        cmd = " ".join([self._nrpe_script, self._nrpe_opts])
        nrpe_setup.remove_check(shortname=self._shortname, check_cmd=cmd)
        nrpe_setup.write()
        hookenv.log(
            "removed NRPE check for tool [{}]".format(self._shortname), hookenv.DEBUG
        )

    def install(self):
        self._install_common_libs()
        self._install_nrpe_plugin()
        hookenv.log("Installed tool [{}]".format(self._shortname))

    def remove(self):
        self.remove_nrpe_check()
        self._remove_nrpe_plugin()
        hookenv.log("Removed tool [{}]".format(self._shortname))

    @classmethod
    def is_series_supported(cls):
        series = lsb_release()["DISTRIB_CODENAME"]

        # BUG(lp#1890652) The following works around xenial layer-apt bug during test

        if (
            hookenv.config("manufacturer") == "test"
            and series == "xenial"  # noqa:W503
            and issubclass(cls, AptVendorTool)  # noqa:W503
        ):
            return False

        return series in cls.SUPPORTED_SERIES

    def chown(self, path, user="nagios", group="nagios"):
        if os.path.exists(path):
            shutil.chown(path, user=user, group=group)

    def get_output_files(self):
        """Which output files this check tool will generate.

        Override in each subclass of Tool.
        """
        return []

    def install_cronjob(self, cron_user="root"):
        assert self._cron_script is not None

        # Copy the cronjob script to the nagios plugins directory
        src = os.path.join(self._cron_script_dir, self._cron_script)
        dst = shutil.copy(src, self.NRPE_PLUGINS_DIR)
        os.chmod(dst, self.CRONJOB_SCRIPT_MODE)
        os.chown(dst, uid=self.CRONJOB_SCRIPT_UID, gid=self.CRONJOB_SCRIPT_GID)
        hookenv.log(
            "Cronjob script [{}] copied to {}"
            "".format(self._cron_script, self.NRPE_PLUGINS_DIR),
            hookenv.DEBUG,
        )

        cmdline = [dst]

        if self._cron_script_args and isinstance(self._cron_script_args, str):
            cmdline.extend([shlex.quote(arg) for arg in self._cron_script_args.split()])
        elif hookenv.config("manufacturer") != "test":
            # Run it once to generate the temp file unless we're on a test
            # container, otherwise the nrpe check # might fail at first.
            # For security reasons, cronjobs that allow parameters shared
            # by the user don't run the script as root
            # e.g. check_ipmi allows a missing output file
            subprocess.call(cmdline)

        # Generate random cronjob execution (internal in minutes)
        cron_interval = 5
        minutes_offsets = []
        minute_num = binascii.crc_hqx("".join(cmdline).encode(), 0) % cron_interval

        while minute_num < 60:
            minutes_offsets.append(str(minute_num))
            minute_num += cron_interval
        cronjob_line = "{minutes} * * * * {user} {cmd}\n".format(
            minutes=",".join(minutes_offsets), user=cron_user, cmd=" ".join(cmdline)
        )

        crond_file = os.path.join(self.CROND_DIR, "hwhealth_{}".format(self._shortname))
        with open(crond_file, "w") as crond_fd:
            crond_fd.write(cronjob_line)
            hookenv.log("Cronjob configured at {}".format(crond_file), hookenv.DEBUG)

        # ensure for cron output files are owned by nagios
        # NOTE: to write xxx.out.tmp file, dir CRONJOB_OUTPUT_DIR must also be owned
        # by nagios, which should already be done in nrpe charm.
        for filepath in self.get_output_files():
            self.chown(filepath)

        return dst

    def _remove_cronjob(self):
        assert self._cron_script is not None

        crond_file = os.path.join(self.CROND_DIR, "hwhealth_{}".format(self._shortname))
        cron_script = os.path.join(self.NRPE_PLUGINS_DIR, self._cron_script)

        for filename in (crond_file, cron_script):
            if not os.path.exists(filename):
                continue
            os.remove(filename)
        hookenv.log(
            "Removed cronjob files [{}, {}]".format(crond_file, cron_script),
            hookenv.DEBUG,
        )

    def _remove_sudoer(self):
        sudoer_path = Path(self.SUDOERS_DIR) / self._sudoer_file

        if not sudoer_path.exists():
            return
        sudoer_path.unlink()
        hookenv.log("deleted sudoer file: {}".format(sudoer_path), hookenv.DEBUG)


class VendorTool(Tool):
    """A class representing a vendor tool binary that is shipped via resource.

    A Vendor tool is a statically linked binary we ship via a resource.
    Since it generally requires root privileges to run, we also deploy a
    cronjob that runs as root and saves the tool output in a temporary file
    that nrpe can read (as nagios user).
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.checksums = []

    def install(self):
        self._install_from_resource()
        self.install_cronjob()
        super().install()

    def remove(self):
        self._remove_cronjob()
        self._remove_binary()
        super().remove()

    def _install_from_resource(self):
        resource = hookenv.resource_get("tools")

        if not resource:
            raise JujuResourceNotFound("tools")
        else:
            hookenv.log(
                "Installing tool [{}] from resource".format(self._shortname),
                hookenv.DEBUG,
            )
            # Move in from a temp directory to be atomic
            with TemporaryDirectory() as tmpdir:
                try:
                    with ZipFile(resource, "r") as zipfile:
                        tmpfile = zipfile.extract(self._shortname, tmpdir)
                        # Verify checksum
                        checksum = hashlib.sha256()
                        with open(tmpfile, "rb") as fd:
                            checksum.update(fd.read())

                        if checksum.hexdigest() not in self.checksums:
                            checksums_string = ", ".join(self.checksums)
                            raise ToolChecksumError(
                                {
                                    "shortname": self._shortname,
                                    "checksum": checksum.hexdigest(),
                                    "expected_checksums": checksums_string,
                                }
                            )
                        # We could just use self.TOOLS_DIR as a destination
                        # here, but shutil.move refuses to overwrite the
                        # destination file unless it receives a full path
                        dst = os.path.join(self.TOOLS_DIR, self._shortname)
                        shutil.move(tmpfile, dst)
                        os.chmod(dst, self.TOOLS_MODE)
                        os.chown(dst, uid=self.TOOLS_UID, gid=self.TOOLS_GID)

                except BadZipFile as error:
                    hookenv.log("BadZipFile: {}".format(error), hookenv.ERROR)

                except PermissionError as error:
                    hookenv.log(
                        "Unable to unzip tool {} "
                        "from the provided resource: {}"
                        "".format(self._shortname, error),
                        hookenv.ERROR,
                    )
                except KeyError as error:
                    raise ToolNotFound({"shortname": self._shortname, "error": error})

    def _remove_binary(self):
        binary_path = Path(self.TOOLS_DIR) / self._shortname

        if not binary_path.exists():
            return
        binary_path.unlink()
        hookenv.log("Removed binary tool {}".format(binary_path), hookenv.DEBUG)


class AptVendorTool(Tool):
    """A class representing a vendor tool binary package available via apt.

    An AptVendorTool is a statically linked binary or set of binaries,
    config files, and libraries available via apt repository from Ubuntu
    or Hardware Vendor package Archives. Since it requires root privileges
    to run locally, we also deploy a cronjob that runs as root and saves
    the tool output in a temporary file that nrpe can read (as nagios user).

    """

    HPE_ILOREST_KEY = """-----BEGIN PGP PUBLIC KEY BLOCK-----
    Version: GnuPG v1.4.12 (GNU/Linux)

    mQENBFZp0LkBCACXajRw3b4x7G7dulNYj0hUID4BtVFq/MjEb6PHckTxGxZDoQRX
    RK54tiTFA9wq3b4P3yEFnOjbjRoI0d7Ls67FADugFO+cDCtsV9yuDlaYP/U/h2nX
    N0R4AdYbsVd5yr6xr+GAy66Hmx5jFH3kbC+zJpOcI0tU9hcyU7gjbxu6KQ1ypI2Q
    VRKf8sRBJXgmkOlbYx35ZUMFcmVxrLJXvUuxmAVXgT9f5M3Z3rsGt/ab+/+1TFSb
    RsaqHsIPE0QH8ikqW4IeDQAo1T99pCdf7FWr45KFFTo7O4AZdLMWVgqeFHaSoZxJ
    307VIINsWiwQoPp0tfU5NOOOwB1Sv3x9QgFtABEBAAG0P0hld2xldHQgUGFja2Fy
    ZCBFbnRlcnByaXNlIENvbXBhbnkgUlNBLTIwNDgtMjUgPHNpZ25ocEBocGUuY29t
    PokBPQQTAQIAJwUCVmnQuQIbLwUJEswDAAYLCQgHAwIGFQgCCQoLAxYCAQIeAQIX
    gAAKCRDCCK3eJsK3l9G+B/0ekblsBeN+xHIJ28pvo2aGb2KtWBwbT1ugI+aIS17K
    UQyHZJUQH+ZeRLvosuoiQEdcGIqmOxi2hVhSCQAOV1LAonY16ACveA5DFAEBz1+a
    WQyx6sOLLEAVX1VqGlBXxh3XLEUWOhlAf1gZPNtHsmURTUy2h1Lv/Yoj8KLyuK2n
    DmrLOS3Ro+RqWocaJfvAgXKgt6Fq/ChDUHOnar7lGswzMsbE/yzLJ7He4y89ImK+
    2ktR5HhDuxqgCe9CWH6Q/1WGhUa0hZ3nbluq7maa+kPe2g7JcRzPH/nJuDCAOZ7U
    6mHE8j0kMQMYjgaYEx2wc02aQRmPyxhbDLjSbtjomXRr
    =voON
    -----END PGP PUBLIC KEY BLOCK-----"""

    HPE_MCP_KEY = """-----BEGIN PGP PUBLIC KEY BLOCK-----
    Version: GnuPG v1.4.12 (GNU/Linux)

    mQENBFZp0LkBCACXajRw3b4x7G7dulNYj0hUID4BtVFq/MjEb6PHckTxGxZDoQRX
    RK54tiTFA9wq3b4P3yEFnOjbjRoI0d7Ls67FADugFO+cDCtsV9yuDlaYP/U/h2nX
    N0R4AdYbsVd5yr6xr+GAy66Hmx5jFH3kbC+zJpOcI0tU9hcyU7gjbxu6KQ1ypI2Q
    VRKf8sRBJXgmkOlbYx35ZUMFcmVxrLJXvUuxmAVXgT9f5M3Z3rsGt/ab+/+1TFSb
    RsaqHsIPE0QH8ikqW4IeDQAo1T99pCdf7FWr45KFFTo7O4AZdLMWVgqeFHaSoZxJ
    307VIINsWiwQoPp0tfU5NOOOwB1Sv3x9QgFtABEBAAG0P0hld2xldHQgUGFja2Fy
    ZCBFbnRlcnByaXNlIENvbXBhbnkgUlNBLTIwNDgtMjUgPHNpZ25ocEBocGUuY29t
    PokBPQQTAQIAJwUCVmnQuQIbLwUJEswDAAYLCQgHAwIGFQgCCQoLAxYCAQIeAQIX
    gAAKCRDCCK3eJsK3l9G+B/0ekblsBeN+xHIJ28pvo2aGb2KtWBwbT1ugI+aIS17K
    UQyHZJUQH+ZeRLvosuoiQEdcGIqmOxi2hVhSCQAOV1LAonY16ACveA5DFAEBz1+a
    WQyx6sOLLEAVX1VqGlBXxh3XLEUWOhlAf1gZPNtHsmURTUy2h1Lv/Yoj8KLyuK2n
    DmrLOS3Ro+RqWocaJfvAgXKgt6Fq/ChDUHOnar7lGswzMsbE/yzLJ7He4y89ImK+
    2ktR5HhDuxqgCe9CWH6Q/1WGhUa0hZ3nbluq7maa+kPe2g7JcRzPH/nJuDCAOZ7U
    6mHE8j0kMQMYjgaYEx2wc02aQRmPyxhbDLjSbtjomXRr
    =voON
    -----END PGP PUBLIC KEY BLOCK-----"""

    HPE_MCP_REPO_TMPL = (
        "deb http://downloads.linux.hpe.com/SDR/repo/mcp {series}/current-gen9 non-free"
    )
    HPE_ILOREST_REPO_TMPL = (
        "deb http://downloads.linux.hpe.com/SDR/repo/ilorest {series}/current non-free"
    )

    # HP doesn't have focal APT sources as of yet
    APT_SOURCES = {
        "ssacli": {
            "xenial": HPE_MCP_REPO_TMPL.format(series="xenial"),
            "bionic": HPE_MCP_REPO_TMPL.format(series="bionic"),
        },
        "ilorest": {
            "xenial": HPE_ILOREST_REPO_TMPL.format(series="xenial"),
            "bionic": HPE_ILOREST_REPO_TMPL.format(series="bionic"),
        },
        "hplog": {
            "xenial": HPE_MCP_REPO_TMPL.format(series="xenial"),
            "bionic": HPE_MCP_REPO_TMPL.format(series="bionic"),
        },
    }
    APT_KEYS = {
        "ssacli": {"xenial": HPE_MCP_KEY, "bionic": HPE_MCP_KEY},
        "ilorest": {"xenial": HPE_ILOREST_KEY, "bionic": HPE_ILOREST_KEY},
        "hplog": {"xenial": HPE_MCP_KEY, "bionic": HPE_MCP_KEY},
    }

    def __init__(self, shortname=None, apt_packages=[]):
        super().__init__(
            shortname=shortname,
            nrpe_script="check_hw_health_cron_output.py",
        )
        self._nrpe_opts = "--filename {}/{}.out".format(
            self.CRONJOB_OUTPUT_DIR, self._shortname
        )
        self.apt_packages = (
            apt_packages if apt_packages else [self.__class__.__name__.lower()]
        )
        self._nrpe_script_dir = self._common_libs_dir

    def install(self):
        self._install_packages()
        self.install_cronjob()
        super().install()

    def remove(self):
        self._remove_cronjob()
        self._remove_packages()
        super().remove()

    def _install_packages(self):
        """Install the apt packages via layer-apt.

        As deb packages manage their own checksums and each cloud can point to
        trusted repositories via install_sources and install_keys config, no
        checksumming is necessary as in resource-shipped VendorTool.

        Each instantiation should configure it's apt_packages based on the
        hardware present on the system.
        """
        self._add_apt_source()

        if hookenv.config("manufacturer") == "test":
            # If we are forcing install on a container for functional tests,
            # we should only download and not install the packages, as some
            # vendor tools depend on hardware to be preset to complete postinst
            # need one option added per package
            apt.queue_install(
                self.apt_packages,
                options=["--download-only" for _ in self.apt_packages],
            )
        else:
            apt.queue_install(self.apt_packages)

    def _add_apt_source(self):
        series = lsb_release()["DISTRIB_CODENAME"]

        if (
            self._shortname not in self.APT_SOURCES
            and self._shortname not in self.APT_KEYS  # noqa:W503
        ):
            return

        if series in self.APT_SOURCES[self._shortname]:
            apt.add_source(
                self.APT_SOURCES[self._shortname][series],
                key=self.APT_KEYS[self._shortname][series],
            )

    def _remove_packages(self):
        apt.purge(self.apt_packages)

    def install_cronjob(self):
        hookenv.log(
            "Attempting AptVendorTool cronjob script install [{}]"
            "".format(self._cron_script),
            hookenv.DEBUG,
        )
        # Don't install a cronjob until the tools are installed

        if self.is_apt_installed():
            hookenv.log(
                "calling install_cronjob for {}".format(self._cron_script),
                hookenv.DEBUG,
            )
            super().install_cronjob()

    def configure_nrpe_check(self, nrpe_setup):
        # Don't drop nrpe checks until the APT package is successfully
        # installed or we'll get red warnings due to Apt errors which
        # should just be reported through juju status

        if self.is_apt_installed():
            super().configure_nrpe_check(nrpe_setup)

    def is_apt_installed(self):
        """Check if packaged queued for install are installed.

        The apt layer will install packages in the background, this method
        allows the reactive layer to know when the apt packages have finished
        installing so that it may go on to configure the nrpe layer of tools
        """

        if hookenv.config("manufacturer") == "test":
            # it is okay to skip this part in testing as layer-apt will block
            # if the package is not downloadable from sources

            return True
        apt_installed_packages = apt.installed()

        for package in self.apt_packages:
            if package not in apt_installed_packages:
                return False

        return True


class Sas3Ircu(VendorTool):
    """A class representing the sas3ircu tool.

    This is a tool supporting the LSI SAS 12Gb/s controllers
    """

    def __init__(self):
        super().__init__(cron_script="cron_sas3ircu.sh")
        self.checksums = [
            "f150eb37bb332668949a3eccf9636e0e03f874aecd17a39d586082c6be1386bd",
            "d69967057992134df1b136f83bc775a641e32c4efc741def3ef6f6a25a9a14b5",
        ]


class Sas2Ircu(VendorTool):
    """A class representing the sas2ircu tool.

    This is a tool supporting the LSI SAS 6Gb/s controllers
    """

    def __init__(self):
        super().__init__(cron_script="cron_sas2ircu.sh")
        self.checksums = [
            "37467826d0b22aad47287efe70bb34e47f475d70e9b1b64cbd63f57607701e73"
        ]  # noqa: E501


class MegaCLI(VendorTool):
    """A class representing the megacli tool.

    This is a tool supporting the LSI MegaRAID SAS controllers
    """

    def __init__(self):
        super().__init__(cron_script="cron_megacli.sh")
        self.checksums = [
            "34f1a235543662615ee35f458317380b3f89fac0e415dee755e0dbc7c4cf6f92",
            "1c4effe33ee5db82227e05925dd629771fd49c7d2be2382d48c48a864452cdec",
            "1a68e6646d1e3dfb7039f581be994500d0ed02de2f928e57399e86473d4c8662",
        ]


class HpLog(AptVendorTool):
    """A class representing the hplog tool.

    This is a tool supporting the LSI MegaRAID SAS controllers
    """

    SUPPORTED_SERIES = ["xenial", "bionic"]

    def __init__(self):
        super().__init__(apt_packages=["hp-health"])


class SsaCli(AptVendorTool):
    """A class representing the ssacli tool.

    This is a tool supporting the HP Smart Array controllers
    """

    SUPPORTED_SERIES = ["xenial", "bionic"]

    def __init__(self):
        super().__init__()


class ILOrest(AptVendorTool):
    """A class representing the ILOrest vendor tool (HPE hardware (Gen 10+)."""

    SUPPORTED_SERIES = ["xenial", "bionic"]

    def __init__(self):
        super().__init__()


class Mdadm(VendorTool):
    """A class representing the mdadm tool (software RAID).

    Our mdadm check kind of behaves like a VendorTool for the purpose of
    installation as it has a cronjob + check script
    """

    def __init__(self):
        super().__init__()

    def install(self):
        # mdadm should already be installed, but let's check
        fetch.apt_install(["mdadm"], fatal=True)
        self.install_cronjob()
        # No vendor binary to install
        Tool.install(self)

    def remove(self):
        self._remove_cronjob()
        # There's no vendor binary to remove
        Tool.remove(self)


class Ipmi(Tool):
    """A class representing the check_ipmi_sensor nagios plugin.

    This class represents the check_ipmi_sensor nagios plugin developed by
    Thomas Krenn AG (https://github.com/thomas-krenn/check_ipmi_sensor_v3).
    We extend the Tool class directly here since there's no vendor binary to
    install; the plugin relies only on freeipmi, a few perl modules, and the
    actual nrpe check, which is imported as a git submodule
    """

    # must match cron_ipmi_sensors.py and check_ipmi.py
    OUTPUT_FILE = "/var/lib/nagios/ipmi_sensors.out"
    EXCLUDE_FILE = "/var/lib/nagios/ipmi_exclude"
    SEL_DATE_FILTER_START_KEY = "ipmi_sel_date_filter_start"

    def __init__(self, nrpe_opts=""):
        self.set_exclusions(hookenv.config("ipmi_check_exclusions"))
        options = hookenv.config("ipmi_check_options")
        options += " --sexclude %s" % self.EXCLUDE_FILE
        super().__init__(cron_script="cron_ipmi_sensors.py", cron_script_args=options)
        self._sudoer_file = "99-check_ipmi_sensor"

    def get_output_files(self):
        return [
            self.OUTPUT_FILE,
            self.EXCLUDE_FILE,
        ]

    def set_exclusions(self, cfg_exclude_lines):
        try:
            excluded_lns = yaml.safe_load(cfg_exclude_lines)
        except yaml.YAMLError as err:
            raise ValueError("ipmi_check_exclusions is invalid yaml: {}".format(err))

        if not isinstance(excluded_lns, list):
            raise ValueError("ipmi_check_exclusions isn't a yaml list")
        write_file(
            self.EXCLUDE_FILE,
            "\n".join(excluded_lns),
            owner="root",
            group="root",
            perms=0o444,
        )

    def add_date_range_to_cron_args(self, date):
        seloptions_argument = ""
        split_cron_args = self._cron_script_args.split()
        if "--seloptions" in split_cron_args:
            # NOTE: This will always point inside the array since the tool's
            #       init code is adding the --sexclude option.
            seloptions_argument_index = split_cron_args.index("--seloptions") + 1
            seloptions_argument = split_cron_args[seloptions_argument_index]
            # This could happen if the user specified the date range option
            # through the ipmi_check_options config. We override it (the action
            # already warned the user about this).
            if "--date-range" in seloptions_argument:
                seloptions_argument = ",".join(
                    [
                        x
                        for x in seloptions_argument.split(",")
                        if "--date-range" not in x
                    ]
                )
            if seloptions_argument:
                seloptions_argument += ","
        else:
            split_cron_args.extend(["--seloptions", ""])
            seloptions_argument_index = len(split_cron_args) - 1
        seloptions_argument += "--date-range=" + date + "-now"
        split_cron_args[seloptions_argument_index] = seloptions_argument
        self._cron_script_args = " ".join(split_cron_args)

    def install_cronjob(self, cron_user="nagios"):
        date_filter = unitdata.kv().get(self.SEL_DATE_FILTER_START_KEY)
        if date_filter:
            self.add_date_range_to_cron_args(date_filter)
            hookenv.log("Added date filter to ipmi sensors cmdline", hookenv.DEBUG)
        super().install_cronjob(cron_user)

    def configure_nrpe_check(self, nrpe_setup):
        # extra options for check_ipmi_sensors Perl script are configured in
        # the cronjob
        self.install_cronjob()
        super().configure_nrpe_check(nrpe_setup)

    def install(self):
        # Install the sudoer file
        self._install_sudoer()
        # Install Perl script called by the (Python) cronjob
        self._install_nrpe_helper_plugin()
        self.install_cronjob()

        # Install the Python script called by check_nrpe
        super().install()

    def remove(self):
        self._remove_sudoer()
        self._remove_nrpe_helper_plugin()
        self._remove_cronjob()
        super().remove()

    def _install_nrpe_helper_plugin(self):
        original_nrpe_script = self._nrpe_script
        self._nrpe_script = "check_ipmi_sensor"
        super()._install_nrpe_plugin()
        self._nrpe_script = original_nrpe_script

    def _remove_nrpe_helper_plugin(self):
        original_nrpe_script = self._nrpe_script
        self._nrpe_script = "check_ipmi_sensor"
        super()._remove_nrpe_plugin()
        self._nrpe_script = original_nrpe_script

    def _install_sudoer(self):
        # User nagios is normally not able to run ipmi-sel and ipmi-sensors, so
        # we need to add a sudoer entry for it
        src = os.path.join(self._files_dir, self._sudoer_file)
        dst = shutil.copy(src, self.SUDOERS_DIR)
        os.chmod(dst, self.SUDOERS_MODE)
        os.chown(dst, uid=self.SUDOERS_UID, gid=self.SUDOERS_GID)
        hookenv.log(
            "sudoer file for tool [{}] installed at {}".format(self._shortname, dst),
            hookenv.DEBUG,
        )

        return dst


class IpmiSEL(Tool):
    """A class representing the ipmiseld utility.

    This class represents the ipmiseld utility from the freeipmi package.
    ipmiseld is able to forward hardware system event log entries to the host
    Operating System's syslog for logging and potentially forwarding alongside
    regular system logs.
    """

    def install(self):
        """Install the ipmiseld package."""
        # Create cachedir to work around LP#1912347
        cache_dir = Path("/var/cache/ipmiseld")
        cache_dir.mkdir(parents=True, exist_ok=True)
        fetch.apt_install(["freeipmi-ipmiseld"], fatal=True)

    def remove(self):
        fetch.apt_purge(["freeipmi-ipmiseld"], fatal=True)

    def configure_nrpe_check(self, nrpe_setup):
        pass


class Nvme(Tool):
    """A class representing the nvme tool (userspace tooling to control NVMe).

    This is a direct subclass of Tool because unlike a VendorTool we are not
    using a cronjob script
    """

    def __init__(self):
        super().__init__()
        self._sudoer_template = "99-check_nvme.tmpl"
        self._sudoer_file = "99-check_nvme"
        self._cron_script = None

    def install(self):
        # mdadm should already be installed, but let's check
        fetch.apt_install(["nvme-cli"], fatal=True)
        self._render_sudoer()
        super().install()

    def remove(self):
        self._remove_sudoer()
        super().remove()

    def _render_sudoer(self):
        # User nagios is normally not able to run nvme, so
        # we need to add a sudoer entry for the smart-log check only
        src = os.path.join(self._shortname, self._sudoer_template)
        dst = os.path.join(self.SUDOERS_DIR, self._sudoer_file)

        devices = self.__get_nvme_devices()

        if not devices:
            return

        devices = dict(
            [("CHECK{}".format(dev.replace("/", "_").upper()), dev) for dev in devices]
        )
        ctxt = {"devices": devices}
        ctxt["devices_cmnd_aliases"] = ", ".join(devices.keys())

        render(
            source=src,
            target=dst,
            context=ctxt,
            perms=self.SUDOERS_MODE,
            templates_dir=None,
        )
        hookenv.log(
            "sudoer file for tool [{}] installed at {}".format(self._shortname, dst),
            hookenv.DEBUG,
        )

        return dst

    def __get_nvme_devices(self):
        devices = []

        for device in glob.glob("/dev/nvme*"):
            nvme_re = re.match(r"^/dev/nvme\d+$", device)

            if not nvme_re:
                continue
            devices.append(nvme_re.group())

        return devices
