import asyncio
import os
import subprocess
import sys
from datetime import datetime
from os.path import abspath, dirname

sys.path.append("lib")

from hwhealth import hwdiscovery  # noqa: E402
from hwhealth import tools  # noqa: E402

import pytest  # noqa: E402


# Treat all tests as coroutines
pytestmark = pytest.mark.asyncio
SERIES = [
    "focal",
    "bionic",
    "xenial",
]
CHARM_DIR = dirname(dirname(dirname(abspath(__file__))))
CHARM_BUILD_DIR = dirname(CHARM_DIR)
NRPECFG_DIR = "/etc/nagios/nrpe.d"
DEF_TIMEOUT = 600
# These go along with the hpe repos for the hp* tools

################
# Helper funcs #
################


async def deploy_hwhealth_res(model, app_name, res_filename):
    # Attaching resources is not implemented yet in libjuju
    # see https://github.com/juju/python-libjuju/issues/294
    tools_res_path = os.path.join(CHARM_BUILD_DIR, res_filename)
    subprocess.check_call(
        [
            "juju",
            "deploy",
            "-m",
            model.info.name,
            os.path.join(CHARM_BUILD_DIR, "hw-health"),
            app_name,
            "--resource",
            "tools={}".format(tools_res_path),
        ]
    )


async def update_hwhealth_res(model, app_name, res_filename):
    tools_res_path = os.path.join(CHARM_BUILD_DIR, res_filename)
    subprocess.check_call(
        [
            "juju",
            "attach-resource",
            "-m",
            model.info.name,
            app_name,
            "tools={}".format(tools_res_path),
        ]
    )


###################
# Custom fixtures #
###################


@pytest.fixture(scope="module", params=SERIES)
async def deploy_app(request, model):
    """Deploy the hw-health charm as a subordinate of ubuntu."""
    # TODO: this might look nicer if we deployed a bundle instead. It could be
    # a jinja template to handle the parametrization
    release = request.param
    channel = "stable"
    hw_health_app_name = "hw-health-{}".format(release)
    hw_health_checksum_app_name = "hw-health-checksum-{}".format(release)

    for principal_app in ["ubuntu", "nagios"]:
        relname = series = release
        if principal_app == "nagios" and release == "focal":
            # NOTE(aluria): cs:nagios was not available in focal
            # On the other hand, bionic testing would create nagios-bionic
            # hence nagios-bionic2
            relname = "bionic2"
            series = "bionic"
        await model.deploy(
            principal_app,
            application_name="{}-{}".format(principal_app, relname),
            series=series,
            channel=channel,
        )
    await model.deploy(
        "ubuntu",
        application_name="ubuntu-checksum-{}".format(release),
        series=release,
        channel=channel,
    )
    nrpe_app = await model.deploy(
        "nrpe",
        application_name="nrpe-{}".format(release),
        series=release,
        num_units=0,
        channel=channel,
    )
    for ubuntu_unit in ["ubuntu", "ubuntu-checksum"]:
        await nrpe_app.add_relation(
            "general-info", "{}-{}:juju-info".format(ubuntu_unit, release)
        )
    await nrpe_app.add_relation("monitors", "nagios-{}:monitors".format(relname))

    # Attaching resources is not implemented yet in libjuju
    # see https://github.com/juju/python-libjuju/issues/294
    await deploy_hwhealth_res(model, hw_health_app_name, "tools.zip")
    await deploy_hwhealth_res(model, hw_health_checksum_app_name, "tools-checksum.zip")

    # This is pretty horrible, but we can't deploy via libjuju
    while True:
        try:
            hw_health_app = model.applications[hw_health_app_name]
            hw_health_checksum_app = model.applications[hw_health_checksum_app_name]
            break
        except KeyError:
            await asyncio.sleep(5)

    await hw_health_app.add_relation(
        "general-info", "ubuntu-{}:juju-info".format(release)
    )
    await hw_health_app.add_relation(
        "nrpe-external-master", "{}:nrpe-external-master".format(nrpe_app.name)
    )

    await hw_health_checksum_app.add_relation(
        "general-info", "ubuntu-checksum-{}:juju-info".format(release)
    )
    await hw_health_checksum_app.add_relation(
        "nrpe-external-master", "{}:nrpe-external-master".format(nrpe_app.name)
    )

    # The app will initially be in blocked state because it's running in a
    # container
    await model.block_until(
        lambda: (
            hw_health_app.status == "blocked"
            and hw_health_checksum_app.status == "blocked"  # noqa W503
        ),
        timeout=DEF_TIMEOUT,
    )
    yield hw_health_app


@pytest.fixture(scope="module")
async def deployed_unit(deploy_app):
    """Return the hw-health unit we've deployed."""
    return deploy_app.units[0]


def _monkey_config(k=None):
    if k is None:
        return {}
    return {"manufacturer": "test"}.get(k)


@pytest.fixture(scope="function")
async def toolset(monkeypatch):
    # All tool classes know which files should be installed and how, so we can
    # use them to read the expected stat results. Monkeypatching is however
    # required as the classes code is not expected to be run outside of a
    # deployed charm

    with monkeypatch.context() as m:
        m.setattr("charmhelpers.core.hookenv.charm_dir", lambda: CHARM_BUILD_DIR)
        m.setattr("charmhelpers.core.hookenv.config", _monkey_config)
        m.setattr(
            "charmhelpers.contrib.charmsupport.nrpe.get_nagios_hostname",
            lambda: "pytest",
        )
        return [tool() for tool in hwdiscovery.get_tools("test")]


#########
# Tests #
#########


async def test_cannot_run_in_container(deploy_app):
    assert deploy_app.status == "blocked"


async def test_forced_deploy(deploy_app, model, run_command):
    # Create a fake NVMe device for the cronjob to be configured
    create_fake_nvme_cmd = "/bin/bash -c 'touch /dev/nvme0'"
    series = deploy_app.name.split("-")[-1]
    for unit in model.units.values():
        if unit.entity_id.startswith("ubuntu-{}".format(series)):
            ubuntu_unit = unit
            await model.block_until(
                lambda: ubuntu_unit.workload_status == "active", timeout=DEF_TIMEOUT
            )
            await run_command(create_fake_nvme_cmd, ubuntu_unit)
            break

    await deploy_app.set_config({"manufacturer": "test"})
    await model.block_until(lambda: deploy_app.status == "active", timeout=DEF_TIMEOUT)
    assert deploy_app.status == "active"


async def test_checksum_forced_deploy(deploy_app, model, run_command):
    # Create a fake NVMe device for the cronjob to be configured
    create_fake_nvme_cmd = "/bin/bash -c 'touch /dev/nvme0'"
    series = deploy_app.name.split("-")[-1]
    checksum_app_name = "hw-health-checksum-{}".format(series)
    checksum_app = model.applications[checksum_app_name]
    for unit in model.units.values():
        if unit.entity_id.startswith("ubuntu-checksum-{}".format(series)):
            ubuntu_unit = unit
            await model.block_until(
                lambda: ubuntu_unit.workload_status == "active", timeout=DEF_TIMEOUT
            )
            await run_command(create_fake_nvme_cmd, ubuntu_unit)
        elif unit.entity_id.startswith(checksum_app_name):
            checksum_unit = unit

    await checksum_app.set_config({"manufacturer": "test"})
    try:
        await model.block_until(
            lambda: (
                checksum_app.status == "blocked"
                and checksum_unit.workload_status_message  # noqa W503
                == "Tool megacli - checksum error"  # noqa E501
            ),
            timeout=DEF_TIMEOUT,
        )
    except asyncio.TimeoutError:
        print(
            "failed to get expected state 'blocked:Tool megacli - checksum error', "
            "witnessed '{}:{}'".format(
                checksum_app.status, checksum_unit.workload_status_message
            )
        )
        assert checksum_app.status == "blocked"
        assert checksum_unit.workload_status_message == "Tool megacli - checksum error"


async def test_checksum_updated_resource_missing(deploy_app, model):
    series = deploy_app.name.split("-")[-1]
    checksum_app_name = "hw-health-checksum-{}".format(series)
    checksum_app = model.applications[checksum_app_name]
    await update_hwhealth_res(model, checksum_app_name, "tools-missing.zip")
    for unit in model.units.values():
        if unit.entity_id.startswith(checksum_app_name):
            checksum_unit = unit
            break

    await model.block_until(
        lambda: (
            checksum_app.status == "blocked"
            and checksum_unit.workload_status_message  # noqa W503
            == "Tool megacli not found"  # noqa E501
        ),
        timeout=DEF_TIMEOUT,
    )


async def test_checksum_updated_resource_ok(deploy_app, model):
    series = deploy_app.name.split("-")[-1]
    checksum_app_name = "hw-health-checksum-{}".format(series)
    checksum_app = model.applications[checksum_app_name]
    await update_hwhealth_res(model, checksum_app_name, "tools.zip")
    for unit in model.units.values():
        if unit.entity_id.startswith(checksum_app_name):
            checksum_unit = unit
            break

    await model.block_until(
        lambda: (
            checksum_app.status == "active"
            and checksum_unit.workload_status_message == "ready"  # noqa W503
        ),
        timeout=DEF_TIMEOUT,
    )


async def test_deployed_file_stats(
    monkeypatch, toolset, deploy_app, deployed_unit, file_stat
):
    # This should really be a parametrized test, but fixtures cannot be used as
    # params value as if they were iterators
    # It should also check for other installed files and differentiate between
    # tool types (e.g. tools.Ipmi does not use a vendor binary)
    series = deploy_app.name.split("-")[-1]
    for tool in toolset:
        # Skip tools that are out of series for the currently deployed application
        with monkeypatch.context() as m:
            m.setattr(
                "hwhealth.tools.lsb_release",
                lambda x=None: {"DISTRIB_CODENAME": series},
            )
            m.setattr(
                "charmhelpers.core.hookenv.config",
                _monkey_config,
            )
            if not tool.is_series_supported():
                print(
                    "Skipping tool {}.  Distribution {} not supported.".format(
                        tool, series
                    )
                )
                continue
        # Have we rendered the nrpe check cfg?
        nrpecfg_path = os.path.join(NRPECFG_DIR, "check_{}.cfg".format(tool._shortname))
        print("Checking {}".format(nrpecfg_path))
        test_stat = await file_stat(nrpecfg_path, deployed_unit)
        assert test_stat["size"] > 0

        # Have we installed the nrpe check script?
        nrpescript_path = os.path.join(
            tool.NRPE_PLUGINS_DIR, os.path.basename(tool._nrpe_script)
        )
        print("Checking {}".format(nrpescript_path))
        test_stat = await file_stat(nrpescript_path, deployed_unit)
        assert test_stat["size"] > 0
        assert test_stat["gid"] == tool.NRPE_PLUGINS_GID
        assert test_stat["uid"] == tool.NRPE_PLUGINS_UID
        assert test_stat["mode"] == oct(tool.NRPE_PLUGINS_MODE)

        # Have we installed any common libs?
        for lib in tool._common_libs:
            lib_path = os.path.join(tool.NRPE_PLUGINS_DIR, os.path.basename(lib))
            print("Checking {}".format(nrpescript_path))
            test_stat = await file_stat(lib_path, deployed_unit)
            assert test_stat["size"] > 0
            assert test_stat["gid"] == tool.NRPE_PLUGINS_GID
            assert test_stat["uid"] == tool.NRPE_PLUGINS_UID
            assert test_stat["mode"] == oct(tool.NRPE_PLUGINS_MODE)

        if isinstance(tool, tools.Ipmi) or isinstance(tool, tools.Nvme):
            # Have we added sudo rights for running freeipmi commands?
            sudoer_path = os.path.join(tool.SUDOERS_DIR, tool._sudoer_file)
            print("Checking {}".format(sudoer_path))
            test_stat = await file_stat(sudoer_path, deployed_unit)
            assert test_stat["size"] > 0
            assert test_stat["gid"] == tool.SUDOERS_GID
            assert test_stat["uid"] == tool.SUDOERS_UID
            assert test_stat["mode"] == oct(tool.SUDOERS_MODE)

        if isinstance(tool, tools.Ipmi) or isinstance(tool, tools.VendorTool):
            # Have we installed the cronjob script helper?
            cron_script_path = os.path.join(tool.NRPE_PLUGINS_DIR, tool._cron_script)
            print("Checking {}".format(cron_script_path))
            test_stat = await file_stat(cron_script_path, deployed_unit)
            assert test_stat["size"] > 0
            assert test_stat["gid"] == tool.CRONJOB_SCRIPT_GID
            assert test_stat["uid"] == tool.CRONJOB_SCRIPT_UID
            assert test_stat["mode"] == oct(tool.CRONJOB_SCRIPT_MODE)

            # Have we installed the cronjob itself?
            cronjob_path = os.path.join(
                tool.CROND_DIR, "hwhealth_{}".format(tool._shortname)
            )
            print("Checking {}".format(cronjob_path))
            test_stat = await file_stat(cronjob_path, deployed_unit)
            assert test_stat["size"] > 0

        if isinstance(tool, tools.VendorTool):
            # Have we installed the vendor binary?
            if isinstance(tool, tools.Mdadm):
                bin_path = os.path.join("/sbin", tool._shortname)
            else:
                bin_path = os.path.join(tool.TOOLS_DIR, tool._shortname)
            print("Checking {}".format(bin_path))
            test_stat = await file_stat(bin_path, deployed_unit)
            assert test_stat["size"] > 0
            assert test_stat["gid"] == tool.TOOLS_GID
            assert test_stat["uid"] == tool.TOOLS_UID
            assert test_stat["mode"] == oct(tool.TOOLS_MODE)


@pytest.mark.parametrize("script_type", ["_nrpe_script", "_cron_script"])
async def test_imports(
    script_type, monkeypatch, toolset, deploy_app, deployed_unit, run_command
):
    """Dry run all auxiliary files to ensure we have all needed dependecies."""
    series = deploy_app.name.split("-")[-1]
    for tool in toolset:
        # Skip tools that are out of series for the currently deployed application
        with monkeypatch.context() as m:
            m.setattr(
                "hwhealth.tools.lsb_release",
                lambda x=None: {"DISTRIB_CODENAME": series},
            )
            m.setattr(
                "charmhelpers.core.hookenv.config",
                _monkey_config,
            )
            if not tool.is_series_supported():
                print(
                    "Skipping tool {}.  Distribution {} not supported.".format(
                        tool, series
                    )
                )
                continue
        script_name = getattr(tool, script_type)
        tool_name = tool.__class__.__name__
        if not script_name:
            # Cannot pytest.skip because it would break out of the loop
            print(
                "Skipping test as {} does not have a {}"
                "".format(tool_name, script_type)
            )
        else:
            print("Checking {}: {}".format(tool_name, script_name))
            path = os.path.join(tool.NRPE_PLUGINS_DIR, script_name)
            cmd = path + " --help"
            results = await run_command(cmd, deployed_unit)
            rc = results["Code"]
            assert rc == "0", "{}, {}. RC is non-zero. results={}".format(
                tool_name, script_name, results
            )


IPMI_CROND_FILE = os.path.join(tools.Ipmi.CROND_DIR, "hwhealth_ipmi")


async def test_action_ack_sel(deployed_unit, run_command):
    action = await deployed_unit.run_action("ack-sel")
    action = await action.wait()

    assert action.status == "completed", "Action did not complete successfully"

    # NOTE: This test might fail if run very very close to UTC midnight... I'm
    # purposedly not contemplating that case in favor of test simplicity.
    today = datetime.utcnow().strftime("%m/%d/%Y")
    cmd = "grep -E '[ ,]--date-range={}-now([ ,]|$)' {}".format(today, IPMI_CROND_FILE)
    result = await run_command(cmd, deployed_unit)
    assert (
        result["Code"] == "0"
    ), "--date-range parameter not found on cron file (or it has the wrong date)"


async def test_action_ack_sel_date_param(deployed_unit, run_command):
    action = await deployed_unit.run_action("ack-sel", date="2021-3-9")
    action = await action.wait()

    assert action.status == "completed", "Action did not complete successfully"

    cmd = "grep -E '[ ,]--date-range=03/09/2021-now([ ,]|$)' {}".format(IPMI_CROND_FILE)
    result = await run_command(cmd, deployed_unit)
    assert (
        result["Code"] == "0"
    ), "--date-range parameter not found on cron file (or it has the wrong date)"


async def test_action_ack_sel_invalid_date_param(deployed_unit, run_command):
    result = await run_command(
        "TMP=`mktemp` && cp {} $TMP && echo $TMP".format(IPMI_CROND_FILE), deployed_unit
    )
    assert result["Code"] == "0", "Error copying original cron file: {}".format(
        result["Stderr"]
    )
    original_crond = result["Stdout"].strip()

    action = await deployed_unit.run_action("ack-sel", date="03/09/2021")
    action = await action.wait()
    assert action.status == "failed", "Action didn't fail"
    assert (
        "Invalid date format" in action.message
    ), "Expected 'Invalid date format' phrase not in action message: {}".format(
        action.message
    )

    result = await run_command(
        "cmp {} {}".format(original_crond, IPMI_CROND_FILE), deployed_unit
    )
    assert result["Code"] == "0", "Cron file changed when it shouldn't have"


async def test_action_ack_sel_warning(model, deploy_app, deployed_unit, run_command):
    await deploy_app.set_config(
        {"ipmi_check_options": "--seloptions --date-range=2/27/2021-now"}
    )
    await model.block_until(
        lambda: deployed_unit.agent_status == "idle", timeout=DEF_TIMEOUT
    )

    action = await deployed_unit.run_action("ack-sel", date="2021-3-9")
    action = await action.wait()

    assert action.status == "completed", "Action did not complete successfully"

    message = action.results["message"]
    # NOTE: We don't check the actual date range value, that's covered on the
    # unit tests and tests above. We just wanna see if the warning's there.
    assert (
        "WARNING: Will override" in message
    ), "Warning about overriding user config is not there: {}".format(message)


async def test_action_unack_sel(model, deploy_app, deployed_unit, run_command):
    await deploy_app.set_config({"ipmi_check_options": ""})
    await model.block_until(
        lambda: deployed_unit.agent_status == "idle", timeout=DEF_TIMEOUT
    )

    cmd = "grep -Fe '--date-range' {}".format(IPMI_CROND_FILE)
    result = await run_command(cmd, deployed_unit)
    assert result["Code"] == "0", "Test expected --date-range to be there but it isn't"

    action = await deployed_unit.run_action("unack-sel")
    action = await action.wait()

    assert action.status == "completed", "Action did not complete successfully"

    cmd = "grep -Fe '--date-range' {}".format(IPMI_CROND_FILE)
    result = await run_command(cmd, deployed_unit)
    assert result["Code"] == "1", "--date-range parameter is still there"


async def test_removal(monkeypatch, toolset, model, deploy_app, file_stat):
    """Remove the unit, test that all files have been cleaned up."""
    hw_health_app_name = deploy_app.name
    series = deploy_app.name.split("-")[-1]
    await deploy_app.remove()
    await model.block_until(
        lambda: hw_health_app_name not in model.applications, timeout=DEF_TIMEOUT
    )
    # Since we've removed the hw-health app, we can't target it anymore, we
    # need to find the principal unit
    for unit in model.units.values():
        if unit.entity_id.startswith("ubuntu-{}".format(series)):
            ubuntu_unit = unit
    for tool in toolset:
        # Skip tools that are out of series for the currently deployed application
        with monkeypatch.context() as m:
            m.setattr(
                "hwhealth.tools.lsb_release",
                lambda x=None: {"DISTRIB_CODENAME": series},
            )
            m.setattr(
                "charmhelpers.core.hookenv.config",
                _monkey_config,
            )
            if not tool.is_series_supported():
                print(
                    "Skipping tool {}.  Distribution {} not supported.".format(
                        tool, series
                    )
                )
                continue
        # Have we removed the nrpe check cfg?
        nrpecfg_path = os.path.join(NRPECFG_DIR, "check_{}.cfg".format(tool._shortname))
        print("Checking {}".format(nrpecfg_path))
        with pytest.raises(RuntimeError):
            await file_stat(nrpecfg_path, ubuntu_unit)

        # Have we removed the nrpe check script?
        nrpescript_path = os.path.join(tool.NRPE_PLUGINS_DIR, tool._nrpe_script)
        print("Checking {}".format(nrpescript_path))
        with pytest.raises(RuntimeError):
            await file_stat(nrpescript_path, ubuntu_unit)

        if isinstance(tool, tools.Ipmi) or isinstance(tool, tools.Nvme):
            # Have we removed sudo rights for running freeipmi commands?
            sudoer_path = os.path.join(tool.SUDOERS_DIR, tool._sudoer_file)
            print("Checking {}".format(sudoer_path))
            with pytest.raises(RuntimeError):
                await file_stat(sudoer_path, ubuntu_unit)

        if isinstance(tool, tools.Ipmi) or isinstance(tool, tools.VendorTool):
            # Have we removed the cronjob script helper?
            cronjob_path = os.path.join(tool.NRPE_PLUGINS_DIR, tool._cron_script)
            print("Checking {}".format(cronjob_path))
            with pytest.raises(RuntimeError):
                await file_stat(cronjob_path, ubuntu_unit)

            # Have we removed the cronjob itself?
            cronjob_path = os.path.join(
                tool.CROND_DIR, "hwhealth_{}".format(tool._shortname)
            )
            print("Checking {}".format(cronjob_path))
            with pytest.raises(RuntimeError):
                await file_stat(cronjob_path, ubuntu_unit)

        if isinstance(tool, tools.VendorTool) and not isinstance(
            tool, tools.Mdadm
        ):  # noqa E501
            # /sbin/mdadm will not be removed, but the vendor binaries
            # should have been
            bin_path = os.path.join(tool.TOOLS_DIR, tool._shortname)
            print("Checking {}".format(bin_path))
            with pytest.raises(RuntimeError):
                await file_stat(bin_path, ubuntu_unit)
