import io
import os
import subprocess
import sys  # noqa: F401
import unittest
import unittest.mock as mock

sys.path.append(os.path.join(os.path.dirname(__file__), "../../files/mdadm"))
import cron_mdadm  # noqa: E402

sys.path.append(os.path.join(os.path.dirname(__file__), "lib"))
from samples import get_sample  # noqa: E402


class TestCronMdadm(unittest.TestCase):
    @mock.patch("os.path.exists")
    @mock.patch("subprocess.check_output")
    def test_get_devices_mdadm_exists(self, mdadm_details, path_exists):
        mdadm_details.return_value = (
            b"ARRAY /dev/md0 metadata=1.2 name=node00:0"
            b" UUID=dfaf7413:7551b000:56dd7442:5b020adb\n"
            b"ARRAY /dev/md1 metadata=1.2 name=node01:0"
            b" UUID=dfaf7413:7551b000:56dd7442:5b020adc\n"
        )
        path_exists.return_value = True
        self.assertEqual(cron_mdadm.get_devices(), set(["/dev/md0", "/dev/md1"]))

    @mock.patch("os.path.exists")
    def test_get_devices_mdadm_notfound(self, path_exists):
        path_exists.return_value = False
        self.assertEqual(cron_mdadm.get_devices(), set())

    @mock.patch("cron_mdadm.generate_output", return_value=False)
    @mock.patch("os.path.exists", return_value=True)
    @mock.patch("subprocess.Popen")
    @mock.patch("sys.exit", return_value=0)
    def test_get_devices_mdadm_exception(
        self, mock_exit, mdadm_details, path_exists, no_genout
    ):
        mdadm_details.side_effect = subprocess.CalledProcessError(1, "unittest")
        self.assertEqual(cron_mdadm.get_devices(), set())

    @mock.patch("cron_mdadm.generate_output")
    @mock.patch("cron_mdadm.get_devices")
    @mock.patch("subprocess.check_output")
    @mock.patch("sys.stdout", new_callable=io.StringIO)
    def test_parse_output_ok(self, mock_print, mdadm_details, devices, genout):
        class TestPopen(object):
            @classmethod
            def __init__(cls):
                test_output = get_sample("mdadm.output")
                cls.stdout = io.FileIO(test_output)
                cls.wait = lambda: 0

        test_output = get_sample("mdadm.output")
        with open(test_output, "r") as fd:
            mdadm_details.return_value = "".join(fd.readlines()).encode()
        devices.return_value = set(["/dev/md0", "/dev/md1", "/dev/md2"])
        genout.return_value = True
        self.assertTrue(cron_mdadm.parse_output())
        genout.assert_called_once_with(
            "OK: /dev/md0 ok; /dev/md1 ok; /dev/md3 ok; /dev/md2 ok"
        )

    @mock.patch("cron_mdadm.generate_output")
    @mock.patch("cron_mdadm.get_devices")
    def test_parse_output_nodevices(self, devices, genout):
        devices.return_value = set()
        genout.return_value = True
        self.assertTrue(cron_mdadm.parse_output())
        genout.assert_called_once_with("WARNING: unexpectedly checked no devices")

    @mock.patch("cron_mdadm.generate_output")
    @mock.patch("cron_mdadm.get_devices")
    @mock.patch("subprocess.check_output")
    def test_parse_output_mdadm_warning(self, mdadm_details, devices, genout):
        devices.return_value = set(["/dev/md0", "/dev/md1", "/dev/md2"])
        mdadm_details.side_effect = subprocess.CalledProcessError(1, "unittest")
        genout.return_value = True
        self.assertTrue(cron_mdadm.parse_output())
        genout.assert_called_once_with(
            "WARNING: error executing mdadm: Command 'unittest'"
            " returned non-zero exit status 1."
        )

    @mock.patch("cron_mdadm.generate_output")
    @mock.patch("cron_mdadm.get_devices")
    @mock.patch("subprocess.check_output")
    def test_parse_output_degraded(self, mdadm_details, devices, genout):
        test_output = get_sample("mdadm.output.critical.1")
        with open(test_output, "r") as fd:
            mdadm_details.return_value = "".join(fd.readlines()).encode()

        devices.return_value = set(["/dev/md0", "/dev/md1", "/dev/md2"])
        genout.return_value = True
        expected = (
            "CRITICAL: /dev/md0 ok; /dev/md1 degraded, Active[1],"
            " Failed[1], Spare[0], Working[1]; /dev/md3 ok;"
            " /dev/md2 ok"
        )
        self.assertTrue(cron_mdadm.parse_output())
        genout.assert_called_once_with(expected)

    @mock.patch("cron_mdadm.generate_output")
    @mock.patch("cron_mdadm.get_devices")
    @mock.patch("subprocess.check_output")
    def test_parse_output_removed(self, mdadm_details, devices, genout):
        test_output = get_sample("mdadm.output.critical.2")
        with open(test_output, "r") as fd:
            mdadm_details.return_value = "".join(fd.readlines()).encode()

        devices.return_value = set(["/dev/md1"])
        genout.return_value = True
        expected = "CRITICAL: /dev/md1 degraded, RaidDevice(s) 2 marked removed"
        self.assertTrue(cron_mdadm.parse_output())
        genout.assert_called_once_with(expected)

    @mock.patch("cron_mdadm.generate_output")
    @mock.patch("cron_mdadm.get_devices")
    @mock.patch("subprocess.check_output")
    def test_parse_output_rebuilding(self, mdadm_details, devices, genout):
        test_output = get_sample("mdadm.output.warning")
        with open(test_output, "r") as fd:
            mdadm_details.return_value = "".join(fd.readlines()).encode()

        devices.return_value = set(["/dev/md0", "/dev/md1", "/dev/md2"])
        genout.return_value = True
        expected = "WARNING: /dev/md1 recovering, /dev/sdk2 rebuilding (13% complete)"
        self.assertTrue(cron_mdadm.parse_output())
        genout.assert_called_once_with(expected)
