import os  # noqa: F401
import subprocess  # noqa: F401
import sys
import unittest
import unittest.mock as mock
from glob import glob

sys.path.append("lib")
import hwhealth.hwdiscovery  # noqa: E402
import hwhealth.tools as tools  # noqa: E402
from hwhealth.discovery.lshw import Hardware  # noqa: E402


class TestGetTools(unittest.TestCase):
    def setUp(self):
        self.samples = glob(
            os.path.join(os.getcwd(), "tests/hw-health-samples/lshw.*.json")
        )
        # These are the tools we expect to be returned as usable, given a
        # specific sample lshw json
        self.EXPECTED_TOOLS = {
            "lshw.dell.01.json": {tools.Nvme, tools.MegaCLI},
            "lshw.dell.02.json": {tools.Nvme},
            "lshw.hp.json": {tools.HpLog, tools.SsaCli, tools.Nvme},
            "lshw.hpe.nonvme.json": {tools.ILOrest},
            "lshw.huawei.json": {tools.Sas2Ircu},
            "lshw.supermicro.notools.json": set(),
            "lshw.supermicro.nvme.json": {tools.Nvme},
            "lshw.supermicro.sas.01.json": {tools.Nvme, tools.Sas3Ircu},
            "lshw.supermicro.sas.02.json": {tools.Nvme, tools.Sas3Ircu},
        }

        # Set up mocks here to avoid burdening the test signature
        def disable_ipmi(value):
            return value != "enable_ipmi"

        patchers = [
            mock.patch.object(
                hwhealth.hwdiscovery.tools.nrpe,
                "get_nagios_hostname",
                lambda: "localhost",
            ),
            mock.patch.object(
                hwhealth.hwdiscovery.hookenv,
                "config",
                lambda x: x not in {"enable_ipmi", "enable_ipmiseld"},
            ),
            mock.patch.object(
                hwhealth.hwdiscovery.hookenv,
                "charm_dir",
                lambda: "/a/completely/fake/dir",
            ),
            # silence debug log output
            mock.patch.object(hwhealth.hwdiscovery.hookenv, "log"),
            # Disable mdadm support
            mock.patch.object(hwhealth.hwdiscovery, "_supports_mdadm", lambda: False),
        ]

        # Start all mocks
        for patcher in patchers:
            patcher.start()

        # Clean up all mocks at the end of the testrun
        self.addCleanup(mock.patch.stopall)

    @mock.patch("hwhealth.discovery.lshw.get_distrib_codename")
    @mock.patch("hwhealth.discovery.lshw.get_platform")
    @mock.patch("hwhealth.hwdiscovery.Hardware", autospec=True)
    def test_get_tools(self, mock_hw, mock_get_platform, mock_get_distrib):
        mock_get_platform.return_value = "ubuntu"
        mock_get_distrib.return_value = "bionic"
        for sample in self.samples:
            mock_hw.return_value = Hardware(filename=sample)
            basename = os.path.basename(sample)
            with self.subTest(msg="Subtest {}".format(basename)):
                actual_classes = hwhealth.hwdiscovery.get_tools()
                expected = self.EXPECTED_TOOLS.get(basename)
                self.assertEqual(actual_classes, expected)


class TestMdadmSupport(unittest.TestCase):
    @mock.patch("charmhelpers.core.hookenv.log")
    @mock.patch("os.path.exists")
    @mock.patch("subprocess.check_output")
    def test_supports_mdadm_true(self, mdadm_details, path_exists, mock_log):
        path_exists.return_value = True
        mdadm_details.return_value = (
            b"ARRAY /dev/md0 metadata=1.2 name=node00:0"
            b" UUID=dfaf7413:7551b000:56dd7442:5b020adb\n"
        )
        self.assertEqual(hwhealth.hwdiscovery._supports_mdadm(), True)

    @mock.patch("os.path.exists")
    def test_supports_mdadm_false(self, path_exists):
        path_exists.return_value = False
        self.assertEqual(hwhealth.hwdiscovery._supports_mdadm(), False)
